import os
import sys
import logging
import traceback

from mindsdb.interfaces.storage.db import session, Log
from mindsdb.utilities.config import Config
from functools import partial

global_config = Config().get_all()
telemtry_enabled = os.getenv('CHECK_FOR_UPDATES', '1').lower() not in ['0', 'false', 'False']

if telemtry_enabled:
    import sentry_sdk
    from sentry_sdk import capture_message, add_breadcrumb
    sentry_sdk.init(
        "https://29e64dbdf325404ebf95473d5f4a54d3@o404567.ingest.sentry.io/5633566",
        traces_sample_rate=0  # Set to `1` to experiment with performance metrics
    )


class LoggerWrapper(object):
    def __init__(self, writer_arr, default_writer_pos):
        self._writer_arr = writer_arr
        self.default_writer_pos = default_writer_pos

    def write(self, message):
        if len(message.strip(' \n')) == 0:
            return
        if 'DEBUG:' in message:
            self._writer_arr[0](message)
        elif 'INFO:' in message:
            self._writer_arr[1](message)
        elif 'WARNING:' in message:
            self._writer_arr[2](message)
        elif 'ERROR:' in message:
            self._writer_arr[3](message)
        else:
            self._writer_arr[self.default_writer_pos](message)

    def flush(self):
        pass


class DbHandler(logging.Handler):
    def __init__(self):
        logging.Handler.__init__(self)
        self.company_id = os.environ.get('MINDSDB_COMPANY_ID', None)

    def emit(self, record):
        self.format(record)
        if (
            len(record.message.strip(' \n')) == 0
            or (record.threadName == 'ray_print_logs' and 'mindsdb-logger' not in record.message)
        ):
            return

        log_type = record.levelname
        source = f'file: {record.pathname} - line: {record.lineno}'
        payload = record.msg

        if telemtry_enabled:
            pass
            # @TODO: Enable once we are sure no sensitive info is being outputed in the logs
            # if log_type in ['INFO']:
            #    add_breadcrumb(
            #        category='auth',
            #        message=str(payload),
            #        level='info',
            #    )
            # Might be too much traffic if we send this for users with slow networks
            # if log_type in ['DEBUG']:
            #    add_breadcrumb(
            #        category='auth',
            #        message=str(payload),
            #        level='debug',
            #    )

        if log_type in ['ERROR', 'WARNING']:
            trace = str(traceback.format_stack(limit=20))
            trac_log = Log(log_type='traceback', source=source, payload=trace, company_id=self.company_id)
            session.add(trac_log)
            session.commit()

            if telemtry_enabled:
                add_breadcrumb(
                    category='stack_trace',
                    message=trace,
                    level='info',
                )
                if log_type in ['ERROR']:
                    capture_message(str(payload))
                if log_type in ['WARNING']:
                    capture_message(str(payload))

        log = Log(log_type=str(log_type), source=source, payload=str(payload), company_id=self.company_id)
        session.add(log)
        session.commit()


def fmt_log_record(log_record):
    return {
        'log_from': 'mindsdb',
        'level': log_record.log_type,
        'context': 'unkown',
        'text': log_record.payload,
        'created_at': str(log_record.created_at).split('.')[0]
    }


def get_logs(min_timestamp, max_timestamp, context, level, log_from, limit):
    logs = session.query(Log).filter(
        Log.company_id == os.environ.get('MINDSDB_COMPANY_ID', None),
        Log.created_at > min_timestamp
    )

    if max_timestamp is not None:
        logs = logs.filter(Log.created_at < max_timestamp)

    if context is not None:
        # e.g. datasource/predictor and assoicated id
        pass

    if level is not None:
        logs = logs.filter(Log.log_type == level)

    if log_from is not None:
        # mindsdb/native/lightwood/all
        pass

    if limit is not None:
        logs = logs.limit(limit)

    logs = [fmt_log_record(x) for x in logs]
    return logs


def initialize_log(config=global_config, logger_name='main', wrap_print=False):
    ''' Create new logger
    :param config: object, app config
    :param logger_name: str, name of logger
    :param wrap_print: bool, if true, then print() calls will be wrapped by log.debug() function.
    '''
    log = logging.getLogger(f'mindsdb.{logger_name}')
    log.propagate = False
    log.setLevel(min(
        getattr(logging, config['log']['level']['console']),
        getattr(logging, config['log']['level']['file'])
    ))

    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    console_handler = logging.StreamHandler()
    console_handler.setLevel(config['log']['level'].get('console', logging.INFO))
    console_handler.setFormatter(formatter)
    log.addHandler(console_handler)

    # db_handler = DbHandler()
    # db_handler.setLevel(config['log']['level'].get('db', logging.WARNING))
    # db_handler.setFormatter(formatter)
    # log.addHandler(db_handler)

    if wrap_print:
        sys.stdout = LoggerWrapper([log.debug, log.info, log.warning, log.error], 1)
        sys.stderr = LoggerWrapper([log.debug, log.info, log.warning, log.error], 3)

    log.error = partial(log.error, exc_info=True)
    return log


def get_log(logger_name=None):
    if logger_name is None:
        return logging.getLogger('mindsdb')
    return logging.getLogger(f'mindsdb.{logger_name}')


log = initialize_log()



import requests
from datetime import datetime
from typing import Dict, List, Optional, Any

import dask
import dill
import sqlalchemy
import pandas as pd
import ludwig as lw
from ludwig.api import LudwigModel
from ludwig.automl import auto_train

from mindsdb_sql import parse_sql
from mindsdb.utilities.log import log
from mindsdb.utilities.config import Config
from mindsdb_sql.parser.ast import Join, Select, Identifier, Constant, Star
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.utils import get_join_input, recur_get_conditionals, get_aliased_columns, default_data_gather
from mindsdb.integrations.libs.storage_handler import SqliteStorageHandler
from mindsdb.integrations.libs.base_handler import BaseHandler, PredictiveHandler, DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse,
    HandlerResponse,
    RESPONSE_TYPE
)
from mindsdb_sql.parser.dialects.mindsdb import (
    CreatePredictor,
    RetrainPredictor,
    DropPredictor,
)


class LudwigHandler(PredictiveHandler):

    type = 'ludwig'

    def __init__(self, name):
        """
        Handler to create and use Ludwig AutoML models from MindsDB.
        """  # noqa
        super().__init__(name)
        self.storage = None
        self.parser = parse_sql
        self.dialect = 'mindsdb'
        self.handler_dialect = 'mysql'
        self.dtypes_to_sql = {
            "Number": sqlalchemy.Integer,
            "Binary": sqlalchemy.Text,
            "Category": sqlalchemy.Text,
            "Bag": sqlalchemy.Text,
            "Set": sqlalchemy.Text,
            "Date": sqlalchemy.DateTime,
            "Text": sqlalchemy.Text,
            "H3": sqlalchemy.Text,
            "Sequence": sqlalchemy.Text,
            "Vector": sqlalchemy.Text,
        }  # TODO audio, image?
        self.is_connected = False

    def connect(self, **kwargs) -> HandlerStatusResponse:
        """ Setup storage and check whether Ludwig is available. """  # noqa
        self.storage = SqliteStorageHandler(context=self.name, config=kwargs['config'])
        result = HandlerStatusResponse(False)
        try:
            import ludwig as lw
            from ludwig.api import LudwigModel
            from ludwig.automl import auto_train
            result.success = True
            self.is_connected = True
        except ImportError as e:
            log.error(f'Error importing Ludwig, {e}!')
            result.error_message = str(e)
        return result

    def disconnect(self):
        self.is_connected = False
        return

    def check_connection(self) -> HandlerStatusResponse:
        return HandlerResponse(self.is_connected)

    def get_tables(self) -> HandlerResponse:
        """ Returns name list of trained models.  """  # noqa
        models = self.storage.get('models')
        r = HandlerResponse(
            RESPONSE_TYPE.TABLE,
            pd.DataFrame(
                list(models.keys()) if models else [],
                columns=['model_name']
            )
        )
        return r

    def get_columns(self, table_name: str) -> HandlerResponse:
        """ For any given model, return the input data types. """  # noqa
        try:
            model = dill.loads(self.storage.get('models').get(table_name)['model'])
            if not model:
                model = LudwigModel()
            cfg = model.config  # json-like
            r = HandlerResponse(
                RESPONSE_TYPE.TABLE,
                pd.DataFrame(
                    [v for v in cfg.values()],
                    columns=[k for k in cfg.keys()]
                )
            )
        except Exception as e:
            log.error(f"Could not get columns for model {table_name}, error: {e}")
            r = HandlerResponse(RESPONSE_TYPE.ERROR)
        return r

    def native_query(self, query: Any) -> HandlerResponse:
        statement = self.parser(query, dialect=self.dialect)
        r = HandlerResponse(True)

        if type(statement) == CreatePredictor:
            model_name = statement.name.parts[-1]

            if model_name in self.get_tables().data_frame.values:
                raise Exception("Error: this model already exists!")

            target = statement.targets[0].parts[-1]
            if statement.order_by:
                raise Exception("Ludwig handler does not support time series tasks yet!")

            # get training data from other integration
            handler = MDB_CURRENT_HANDLERS[str(statement.integration_name)]  # TODO import from mindsdb init
            handler_query = self.parser(statement.query_str, dialect=self.handler_dialect)
            df = default_data_gather(handler, handler_query)

            grace_period = 72
            time_budget = 120
            results = auto_train(
                dataset=df,
                target=target,
                time_limit_s=max(grace_period, time_budget),  # TODO customizable (also, is grace period fixed?)
                tune_for_memory=False
            )
            model = results.best_model

            all_models = self.storage.get('models')
            payload = {
                'stmt': statement,
                'model': dill.dumps(model),
            }
            if all_models is not None:
                all_models[model_name] = payload
            else:
                all_models = {model_name: payload}
            self.storage.set('models', all_models)

        elif type(statement) == RetrainPredictor:
            log.warning('Warning: retraining Ludwig models is not yet supported!')  # TODO: restore this
            return r

        elif type(statement) == DropPredictor:
            to_drop = statement.name.parts[-1]
            models = self.storage.get('models')
            if models:
                del models[to_drop]
                self.storage.set('models', models)
            else:
                raise Exception(f"Can't drop non-existent model {to_drop}")

        else:
            raise Exception(f"Query type {type(statement)} not supported")
        
        return r

    def query(self, query: ASTNode) -> HandlerResponse:
        values = recur_get_conditionals(query.where.args, {})
        model_name, _, _ = self._get_model_name(query)
        model = self._get_model(model_name)
        df = pd.DataFrame.from_dict(values)
        df = self._call_model(df, model)
        r = HandlerResponse(
            RESPONSE_TYPE.TABLE,
            df
        )
        return r

    def join(self, stmt, data_handler, into: Optional[str]) -> HandlerResponse:
        """
        Batch prediction using the output of a query passed to a data handler as input for the model.
        """  # noqa

        model_name, model_alias, model_side = self._get_model_name(stmt)
        data_side = 'right' if model_side == 'left' else 'left'
        model = self._get_model(model_name)
        model_input = get_join_input(stmt, model, [model_name, model_alias], data_handler, data_side)

        # get model output and rename columns
        predictions = self._call_model(model_input, model)
        model_input.columns = get_aliased_columns(list(model_input.columns), model_alias, stmt.targets, mode='input')
        predictions.columns = get_aliased_columns(list(predictions.columns), model_alias, stmt.targets, mode='output')

        if into:
            try:
                dtypes = {}
                for col in predictions.columns:
                    if model.dtype_dict.get(col, False):
                        dtypes[col] = self.dtypes_to_sql.get(col, sqlalchemy.Text)

                data_handler.select_into(into, predictions, dtypes=dtypes)
            except Exception as e:
                print("Error when trying to store the JOIN output in data handler.")

        r = HandlerResponse(
            RESPONSE_TYPE.TABLE,
            predictions
        )
        return r

    def _get_model(self, model_name):
        model = dill.loads(self.storage.get('models')[model_name]['model'])
        return model

    def _call_model(self, df, model):
        predictions = dask.compute(model.predict(df)[0])[0]
        predictions.columns = ['prediction']
        joined = df.join(predictions)
        return joined



from setuptools import setup, find_packages

pkgs, new_links = install_deps()

setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__github__'],
    download_url=about['__pypi__'],
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__email__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=pkgs,
    dependency_links=new_links,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6"
)



import time
from torch.multiprocessing import Process


def periodic_executor(freq, func, args):
    while True:
        func(*args)
        
        time.sleep(freq)

def register(freq, func, args):
    print(f'Registered {func} with freq: {freq}')
    p = Process(target=periodic_executor, args=(freq, func, args))
    p.daemon = True
    p.start()
    #asyncio.get_running_loop().run_in_executor(None, periodic_executor, freq,func,args)



import psycopg
from psycopg.pq import ExecStatus
from pandas import DataFrame

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.libs.base_handler import DatabaseHandler
from mindsdb.utilities.log import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


class PostgresHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the PostgreSQL statements.
    """
    name = 'postgres'

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_args = kwargs.get('connection_data')
        self.dialect = 'postgresql'
        self.database = self.connection_args.get('database')
        if 'database' in self.connection_args:
            self.connection_args['database']
        self.renderer = SqlalchemyRender('postgres')

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Handles the connection to a PostgreSQL database insance.
        """
        if self.is_connected is True:
            return self.connection

        # TODO: Check psycopg_pool
        if self.connection_args.get('dbname') is None:
            self.connection_args['dbname'] = self.database
        args = self.connection_args.copy()

        for key in ['type', 'publish', 'test', 'date_last_update', 'integrations_name', 'database_name', 'id', 'database']:
            if key in args:
                del args[key]
        connection = psycopg.connect(**args, connect_timeout=10)

        self.is_connected = True
        self.connection = connection
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the PostgreSQL database
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                cur.execute('select 1;')
            response.success = True
        except psycopg.Error as e:
            log.error(f'Error connecting to PostgreSQL {self.database}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in PostgreSQL
        :return: returns the records from the current recordset
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                if ExecStatus(cur.pgresult.status) == ExecStatus.COMMAND_OK:
                    response = Response(RESPONSE_TYPE.OK)
                else:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        DataFrame(
                            result,
                            columns=[x.name for x in cur.description]
                        )
                    )
                connection.commit()
            except Exception as e:
                log.error(f'Error running query: {query} on {self.database}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all tabels in PostgreSQL without the system tables information_schema and pg_catalog
        """
        query = """
            SELECT
                table_schema,
                table_name,
                table_type
            FROM
                information_schema.tables
            WHERE
                table_schema NOT IN ('information_schema', 'pg_catalog')
                and table_type in ('BASE TABLE', 'VIEW')
        """
        return self.native_query(query)

    def get_columns(self, table_name):
        query = f"""
            SELECT
                column_name as "Field",
                data_type as "Type"
            FROM
                information_schema.columns
            WHERE
                table_name = '{table_name}'
        """
        return self.native_query(query)



import copy

import duckdb
import numpy as np

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.planner.utils import query_traversal
from mindsdb_sql.parser.ast import (
    Select, Identifier,
    Function, Constant
)

from mindsdb.utilities.log import log


def query_df(df, query, session=None):
    """ Perform simple query ('select' from one table, without subqueries and joins) on DataFrame.

        Args:
            df (pandas.DataFrame): data
            query (mindsdb_sql.parser.ast.Select | str): select query

        Returns:
            pandas.DataFrame
    """

    if isinstance(query, str):
        query_ast = parse_sql(query, dialect='mysql')
    else:
        query_ast = copy.deepcopy(query)

    if isinstance(query_ast, Select) is False \
       or isinstance(query_ast.from_table, Identifier) is False:
        raise Exception(
            "Only 'SELECT from TABLE' statements supported for internal query"
        )

    query_ast.from_table.parts = ['df_table']

    def adapt_query(node, is_table, **kwargs):
        if is_table:
            return
        if isinstance(node, Identifier):
            if len(node.parts) > 1:
                node.parts = [node.parts[-1]]
                return node
        if isinstance(node, Function):
            if node.op.lower() == 'database' and len(node.args) == 0:
                if session is not None:
                    cur_db = session.database
                else:
                    cur_db = None
                return Constant(cur_db)

    query_traversal(query_ast, adapt_query)

    render = SqlalchemyRender('postgres')
    try:
        query_str = render.get_string(query_ast, with_failback=False)
    except Exception as e:
        log.error(
            f"Exception during query casting to 'postgres' dialect. Query: {str(query)}. Error: {e}"
        )
        query_str = render.get_string(query_ast, with_failback=True)

    res = duckdb.query_df(df, 'df_table', query_str)
    result_df = res.df()
    result_df = result_df.replace({np.nan: None})

    new_column_names = {}
    real_column_names = [x[0] for x in res.description()]
    for i, duck_column_name in enumerate(result_df.columns):
        new_column_names[duck_column_name] = real_column_names[i]
    result_df = result_df.rename(
        new_column_names,
        axis='columns'
    )
    return result_df



from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum


class PasswordAnswer(Packet):
    def setup(self, length=0, count_header=1, body=''):
        length = len(body)

        if length == 0:
            return
        self.password = Datum('string<NUL>')
        buffer = body
        buffer = self.password.setFromBuff(buffer)



import time

from flask import request
from flask_restx import Resource
from pandas.core.frame import DataFrame

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import Constant, Identifier
from mindsdb_sql.planner.utils import query_traversal

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.analysis import ns_conf
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE as SQL_RESPONSE_TYPE


@ns_conf.route('/query')
class QueryAnalysis(Resource):
    @ns_conf.doc('post_query_to_analyze')
    def post(self):
        data = request.json
        query = data.get('query')
        context = data.get('context', {})
        limit = data.get('limit')
        if query is None or len(query) == 0:
            return http_error(400, 'Missed query', 'Need provide query to analyze')

        try:
            ast = parse_sql(query)
        except Exception as e:
            return http_error(500, 'Wrong query', str(e))

        if limit is not None:
            ast.limit = Constant(limit)
            query = str(ast)

        mysql_proxy = FakeMysqlProxy(
            company_id=request.company_id,
            user_class=request.user_class
        )
        mysql_proxy.set_context(context)

        try:
            result = mysql_proxy.process_query(query)
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            return http_error(500, 'Error', str(e))

        if result.type == SQL_RESPONSE_TYPE.ERROR:
            return http_error(500, f'Error {result.error_code}', result.error_message)
        if result.type != SQL_RESPONSE_TYPE.TABLE:
            return http_error(500, 'Error', 'Query does not return data')

        column_names = [x['name'] for x in result.columns]
        analysis = request.model_interface.analyse_dataset(
            df=DataFrame(result.data, columns=column_names),
            company_id=None
        )

        query_tables = []

        def find_tables(node, is_table, **kwargs):
            if is_table and isinstance(node, Identifier):
                query_tables.append('.'.join(node.parts))
        query_traversal(ast, find_tables)

        return {
            'analysis': analysis,
            'column_names': column_names,
            'row_count': len(result.data),
            'timestamp': time.time(),
            'tables': query_tables
        }



import os
import sys
import traceback
import tempfile
from pathlib import Path
from typing import Optional
import json
import requests
import traceback

import pandas as pd
from pandas.core.frame import DataFrame
import torch.multiprocessing as mp
import lightwood
from lightwood.api.types import ProblemDefinition, JsonAI
from lightwood import __version__ as lightwood_version

from mindsdb import __version__ as mindsdb_version
import mindsdb.interfaces.storage.db as db
from mindsdb.interfaces.storage.db import Integration, session, Predictor
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.utilities.config import Config
from mindsdb.utilities.functions import mark_process
from mindsdb.utilities.log import log
from mindsdb.utilities.with_kwargs_wrapper import WithKWArgsWrapper
from mindsdb.interfaces.database.integrations import IntegrationController
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


ctx = mp.get_context('spawn')


def create_learn_mark():
    if os.name == 'posix':
        p = Path(tempfile.gettempdir()).joinpath('mindsdb/learn_processes/')
        p.mkdir(parents=True, exist_ok=True)
        p.joinpath(f'{os.getpid()}').touch()


def delete_learn_mark():
    if os.name == 'posix':
        p = Path(tempfile.gettempdir()).joinpath('mindsdb/learn_processes/').joinpath(f'{os.getpid()}')
        if p.exists():
            p.unlink()


def rep_recur(org: dict, ovr: dict):
    for k in ovr:
        if k in org:
            if isinstance(org[k], dict) and isinstance(ovr[k], dict):
                rep_recur(org[k], ovr[k])
            else:
                org[k] = ovr[k]
        else:
            org[k] = ovr[k]


def brack_to_mod(ovr):
    if not isinstance(ovr, dict):
        if isinstance(ovr, list):
            for i in range(len(ovr)):
                ovr[i] = brack_to_mod(ovr[i])
        elif isinstance(ovr, str):
            if '(' in ovr and ')' in ovr:
                mod = ovr.split('(')[0]
                args = {}
                if '()' not in ovr:
                    for str_pair in ovr.split('(')[1].split(')')[0].split(','):
                        k = str_pair.split('=')[0].strip(' ')
                        v = str_pair.split('=')[1].strip(' ')
                        args[k] = v

                ovr = {
                    'module': mod,
                    'args': args
                }
            elif '{' in ovr and '}' in ovr:
                try:
                    ovr = json.loads(ovr)
                except Exception:
                    pass
        return ovr
    else:
        for k in ovr.keys():
            ovr[k] = brack_to_mod(ovr[k])

    return ovr


@mark_process(name='learn')
def run_generate(df: DataFrame, problem_definition: ProblemDefinition, predictor_id: int, json_ai_override: dict = None) -> int:
    json_ai = lightwood.json_ai_from_problem(df, problem_definition)
    if json_ai_override is None:
        json_ai_override = {}

    json_ai_override = brack_to_mod(json_ai_override)
    json_ai = json_ai.to_dict()
    rep_recur(json_ai, json_ai_override)
    json_ai = JsonAI.from_dict(json_ai)

    code = lightwood.code_from_json_ai(json_ai)

    predictor_record = Predictor.query.with_for_update().get(predictor_id)
    predictor_record.json_ai = json_ai.to_dict()
    predictor_record.code = code
    db.session.commit()


@mark_process(name='learn')
def run_fit(predictor_id: int, df: pd.DataFrame) -> None:
    try:
        predictor_record = Predictor.query.with_for_update().get(predictor_id)
        assert predictor_record is not None

        fs_store = FsStore()
        config = Config()

        predictor_record.data = {'training_log': 'training'}
        session.commit()
        predictor: lightwood.PredictorInterface = lightwood.predictor_from_code(predictor_record.code)
        predictor.learn(df)

        session.refresh(predictor_record)

        fs_name = f'predictor_{predictor_record.company_id}_{predictor_record.id}'
        pickle_path = os.path.join(config['paths']['predictors'], fs_name)
        predictor.save(pickle_path)

        fs_store.put(fs_name, fs_name, config['paths']['predictors'])

        predictor_record.data = predictor.model_analysis.to_dict()

        # getting training time for each tried model. it is possible to do
        # after training only
        fit_mixers = list(predictor.runtime_log[x] for x in predictor.runtime_log
                          if isinstance(x, tuple) and x[0] == "fit_mixer")
        submodel_data = predictor_record.data.get("submodel_data", [])
        # add training time to other mixers info
        if submodel_data and fit_mixers and len(submodel_data) == len(fit_mixers):
            for i, tr_time in enumerate(fit_mixers):
                submodel_data[i]["training_time"] = tr_time
        predictor_record.data["submodel_data"] = submodel_data

        predictor_record.dtype_dict = predictor.dtype_dict
        session.commit()
    except Exception as e:
        session.refresh(predictor_record)
        predictor_record.data = {'error': f'{traceback.format_exc()}\nMain error: {e}'}
        session.commit()
        raise e


@mark_process(name='learn')
def run_learn_remote(df: DataFrame, predictor_id: int) -> None:
    try:
        serialized_df = json.dumps(df.to_dict())
        predictor_record = Predictor.query.with_for_update().get(predictor_id)
        resp = requests.post(predictor_record.data['train_url'],
                             json={'df': serialized_df, 'target': predictor_record.to_predict[0]})

        assert resp.status_code == 200
        predictor_record.data['status'] = 'complete'
    except Exception:
        predictor_record.data['status'] = 'error'
        predictor_record.data['error'] = str(resp.text)

    session.commit()


@mark_process(name='learn')
def run_learn(df: DataFrame, problem_definition: ProblemDefinition, predictor_id: int,
              json_ai_override: dict = None) -> None:
    if json_ai_override is None:
        json_ai_override = {}
    try:
        run_generate(df, problem_definition, predictor_id, json_ai_override)
        run_fit(predictor_id, df)
    except Exception as e:
        predictor_record = Predictor.query.with_for_update().get(predictor_id)
        print(traceback.format_exc())

        try:
            exception_type, _exception_object, exception_traceback = sys.exc_info()
            filename = exception_traceback.tb_frame.f_code.co_filename
            line_number = exception_traceback.tb_lineno
            error_message = f'{exception_type.__name__}: {e}, raised at: {filename}#{line_number}'
        except Exception:
            error_message = str(e)

        predictor_record.data = {"error": error_message}
        session.commit()


def run_adjust(name, db_name, from_data, datasource_id, company_id):
    # @TODO: Actually implement this
    return 0


@mark_process(name='learn')
def run_update(name: str, df: DataFrame, company_id: int):
    original_name = name
    name = f'{company_id}@@@@@{name}'

    fs_store = FsStore()
    config = Config()

    try:
        predictor_record = Predictor.query.filter_by(company_id=company_id, name=original_name).first()
        assert predictor_record is not None
        predictor_record.update_status = 'updating'
        session.commit()

        problem_definition = predictor_record.learn_args
        problem_definition['target'] = predictor_record.to_predict[0]

        if 'join_learn_process' in problem_definition:
            del problem_definition['join_learn_process']

        # Adapt kwargs to problem definition
        if 'timeseries_settings' in problem_definition:
            problem_definition['timeseries_settings'] = problem_definition['timeseries_settings']

        if 'stop_training_in_x_seconds' in problem_definition:
            problem_definition['time_aim'] = problem_definition['stop_training_in_x_seconds']

        json_ai = lightwood.json_ai_from_problem(df, problem_definition)
        predictor_record.json_ai = json_ai.to_dict()
        predictor_record.code = lightwood.code_from_json_ai(json_ai)
        predictor_record.data = {'training_log': 'training'}
        session.commit()
        predictor: lightwood.PredictorInterface = lightwood.predictor_from_code(predictor_record.code)
        predictor.learn(df)

        fs_name = f'predictor_{predictor_record.company_id}_{predictor_record.id}'
        pickle_path = os.path.join(config['paths']['predictors'], fs_name)
        predictor.save(pickle_path)
        fs_store.put(fs_name, fs_name, config['paths']['predictors'])
        predictor_record.data = predictor.model_analysis.to_dict()  # type: ignore
        session.commit()

        predictor_record.lightwood_version = lightwood_version
        predictor_record.mindsdb_version = mindsdb_version
        predictor_record.update_status = 'up_to_date'
        session.commit()

    except Exception as e:
        log.error(e)
        predictor_record.update_status = 'update_failed'  # type: ignore
        session.commit()
        return str(e)


class LearnRemoteProcess(ctx.Process):
    deamon = True

    def __init__(self, *args):
        super(LearnRemoteProcess, self).__init__(args=args)

    def run(self):
        run_learn_remote(*self._args)


class LearnProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(LearnProcess, self).__init__(args=args)

    def run(self):
        run_learn(*self._args)


class GenerateProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(GenerateProcess, self).__init__(args=args)

    def run(self):
        run_generate(*self._args)


class FitProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(FitProcess, self).__init__(args=args)

    def run(self):
        run_fit(*self._args)


class AdjustProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(AdjustProcess, self).__init__(args=args)

    def run(self):
        '''
        running at subprocess due to
        ValueError: signal only works in main thread

        this is work for celery worker here?
        '''
        run_adjust(*self._args)


class UpdateProcess(ctx.Process):
    daemon = True

    def __init__(self, *args):
        super(UpdateProcess, self).__init__(args=args)

    def run(self):
        run_update(*self._args)



from .information_schema_datanode import InformationSchemaDataNode
from .integration_datanode import IntegrationDataNode
from .mindsdb_datanode import MindsDBDataNode

__all__ = ['InformationSchemaDataNode', 'IntegrationDataNode', 'MindsDBDataNode']



from flask import request
from flask_restx import Resource, abort

from mindsdb.api.http.namespaces.configs.streams import ns_conf
import mindsdb.interfaces.storage.db as db

STREAM_INTEGRATION_TYPES = ('kafka', 'redis')


def get_streams():
    streams = db.session.query(db.Stream).filter_by(company_id=request.company_id).all()
    streams_as_dicts = []
    for s in streams:
        streams_as_dicts.append({
            'name': s.name,
            'predictor': s.predictor,
            'integration': s.integration,
            'stream_in': s.stream_in,
            'stream_out': s.stream_out,
            'type': s.type,
            'connection': s.connection_info
        })
    return streams_as_dicts


@ns_conf.route('/')
class StreamList(Resource):
    @ns_conf.doc("get_streams")
    def get(self):
        return get_streams()


@ns_conf.route('/<name>')
@ns_conf.param('name', 'Key-value storage stream')
class Stream(Resource):
    @ns_conf.doc("get_stream")
    def get(self, name):
        for stream in get_streams():
            if stream['name'] == name:
                return stream
        return abort(404, f'Stream "{name}" doesn\'t exist')

    @ns_conf.doc("put_stream")
    def put(self, name):
        # Back compatibility with previous endpoint version
        params = request.json.get('params') or request.json
        params_keys = params.keys()
        for param in ['predictor', 'stream_in', 'stream_out']:
            if param not in params_keys:
                return abort(400, 'Please provide "{}"'.format(param))
        if 'integration' not in params_keys and 'connection' not in params_keys:
            return abort(400, "'integration' in case of local installation and 'connection' in case of cloud are required.")

        if 'integration' in params_keys:
            integration = request.integration_controller.get(params['integration'])
            if integration is None:
                return abort(404, 'Integration "{}" doesn\'t exist'.format(params['integration']))

            if integration['type'] not in STREAM_INTEGRATION_TYPES:
                return abort(400, 'Integration "{}" is not of type [{}]'.format(
                    params['integration'],
                    '/'.join(STREAM_INTEGRATION_TYPES)
                ))

        else:
            # cloud
            if 'type' not in params_keys:
                return abort(404, "'type' parameter is required in case of cloud.")
            # because '_' is not allowed in pod name - replace it.
            name = name.replace('_', '-')

        if db.session.query(db.Stream).filter_by(company_id=request.company_id, name=name).first() is not None:
            return abort(404, 'Stream "{}" already exists'.format(name))

        if db.session.query(db.Predictor).filter_by(company_id=request.company_id, name=params['predictor']).first() is None:
            return abort(404, 'Predictor "{}" doesn\'t exist'.format(params['predictor']))

        stream = db.Stream(
            company_id=request.company_id,
            name=name,
            integration=params.get('integration'),
            predictor=params['predictor'],
            stream_in=params['stream_in'],
            stream_out=params['stream_out'],
            anomaly_stream=params.get('anomaly_stream'),
            type=params.get('type'),
            connection_info=params.get('connection'),
            learning_params=params.get('learning_params', {}),
            learning_threshold=params.get('learning_threshold', 0)
        )

        db.session.add(stream)
        db.session.commit()

        return {'success': True}, 200

    @ns_conf.doc("delete_stream")
    def delete(self, name):
        stream = db.session.query(db.Stream).filter_by(company_id=request.company_id, name=name).first()
        if stream is None:
            return abort(404, 'Stream "{}" doesn\'t exist'.format(name))
        db.session.delete(stream)
        db.session.commit()

        return {"success": True}, 200



import json
import copy
from datetime import datetime

from lightwood.api import dtype
import pandas as pd
import numpy as np

from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.api.mysql.mysql_proxy.utilities.functions import get_column_in_case
from mindsdb.utilities.functions import cast_row_types
from mindsdb.utilities.config import Config
from mindsdb.api.mysql.mysql_proxy.utilities import SqlApiException
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow


class NumpyJSONEncoder(json.JSONEncoder):
    """
    Use this encoder to avoid
    "TypeError: Object of type float32 is not JSON serializable"

    Example:
    x = np.float32(5)
    json.dumps(x, cls=NumpyJSONEncoder)
    """
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (np.float, np.float32, np.float64)):
            return float(obj)
        else:
            return super().default(obj)


class MindsDBDataNode(DataNode):
    type = 'mindsdb'

    def __init__(self, model_interface, integration_controller):
        self.config = Config()
        self.model_interface = model_interface
        self.integration_controller = integration_controller

    def get_tables(self):
        models = self.model_interface.get_models()
        tables = []
        for model in models:
            tables.append(TablesRow(TABLE_NAME=model['name']))
        tables.append(TablesRow(TABLE_NAME='predictors'))
        tables.append(TablesRow(TABLE_NAME='databases'))

        return tables

    def has_table(self, table):
        names = [table.TABLE_NAME for table in self.get_tables()]
        return table in names

    def _get_model_columns(self, table_name):
        model = self.model_interface.get_model_data(name=table_name)
        dtype_dict = model.get('dtype_dict')
        if isinstance(dtype_dict, dict) is False:
            return []
        columns = []
        columns += list(dtype_dict.keys())
        predict = model['predict']
        if not isinstance(predict, list):
            predict = [predict]
        columns += [f'{x}_original' for x in predict]
        for col in predict:
            if dtype_dict.get(col) in (dtype.integer, dtype.float):
                columns += [f"{col}_min", f"{col}_max"]
            columns += [f"{col}_confidence"]
            columns += [f"{col}_explain"]
        return columns

    def get_table_columns(self, table):
        if table == 'predictors':
            return ['name', 'status', 'accuracy', 'predict', 'update_status',
                    'mindsdb_version', 'error', 'select_data_query',
                    'training_options']
        if table in ('datasources', 'databases'):
            return ['name', 'database_type', 'host', 'port', 'user']

        columns = []

        if table in [x['name'] for x in self.model_interface.get_models()]:
            columns = self._get_model_columns(table)
            columns += ['when_data', 'select_data_query']

        return columns

    def _select_predictors(self):
        models = self.model_interface.get_models()
        columns = ['name', 'status', 'accuracy', 'predict', 'update_status',
                   'mindsdb_version', 'error', 'select_data_query',
                   'training_options']
        return pd.DataFrame([[
            x['name'],
            x['status'],
            str(x['accuracy']) if x['accuracy'] is not None else None,
            ', '.join(x['predict']) if isinstance(x['predict'], list) else x['predict'],
            x['update'],
            x['mindsdb_version'],
            x['error'],
            '',
            ''   # TODO
        ] for x in models], columns=columns)

    def _select_integrations(self):
        integrations = self.integration_controller.get_all()
        result = []
        for ds_name, ds_meta in integrations.items():
            connection_data = ds_meta.get('connection_data', {})
            result.append([
                ds_name, ds_meta.get('engine'), connection_data.get('host'), connection_data.get('port'), connection_data.get('user')
            ])
        return pd.DataFrame(
            result,
            columns=['name', 'database_type', 'host', 'port', 'user']
        )

    def delete_predictor(self, name):
        self.model_interface.delete_model(name)

    def get_predictors(self, mindsdb_sql_query):
        predictors_df = self._select_predictors()

        try:
            result_df = query_df(predictors_df, mindsdb_sql_query)
        except Exception as e:
            print(f'Exception! {e}')
            return [], []

        # FIXME https://github.com/mindsdb/dfsql/issues/38
        # TODO remove it whem wll be sure query_df do properly casting
        # result_df = result_df.where(pd.notnull(result_df), '')

        return result_df.to_dict(orient='records'), list(result_df.columns)

    def get_integrations(self, mindsdb_sql_query):
        datasources_df = self._select_integrations()
        try:
            result_df = query_df(datasources_df, mindsdb_sql_query)
        except Exception as e:
            print(f'Exception! {e}')
            return [], []
        return result_df.to_dict(orient='records'), list(result_df.columns)

    def query(self, table, columns=None, where=None, where_data=None, order_by=None, group_by=None, integration_name=None, integration_type=None):
        ''' NOTE WHERE statements can be just $eq joined with 'and'
        '''
        if table == 'predictors':
            return self._select_predictors()
        if table == 'datasources':
            return self._select_datasources()

        original_when_data = None
        if 'when_data' in where_data:
            if len(where_data) > 1:
                raise ValueError("Should not be used any other keys in 'where', if 'when_data' used")
            try:
                original_when_data = where_data['when_data']
                where_data = json.loads(where_data['when_data'])
                if isinstance(where_data, list) is False:
                    where_data = [where_data]
            except Exception:
                raise ValueError(f'''Error while parse 'when_data'="{where_data}"''')

        select_data_query = None

        new_where = {}
        if where_data is None:
            for key, value in where_data.items():
                if isinstance(value, dict) is False or len(value.keys()) != 1 or list(value.keys())[0] != '$eq':
                    # TODO value should be just string or number
                    raise Exception()
                new_where[key] = value['$eq']

            if len(new_where) == 0:
                return []

            where_data = [new_where]

        if isinstance(where_data, dict):
            where_data = [where_data]

        models_sql_meta = self.get_tables()
        if table not in [x.TABLE_NAME for x in models_sql_meta]:
            raise SqlApiException(f"Predictor '{table}' does not exists'")

        model = self.model_interface.get_model_data(name=table)
        if model.get('status') != 'complete':
            raise SqlApiException(
                f"Trying to use predictor '{table}', but the predictor has status '{model.get('status')}'. " +
                "The predictor must has status 'complete' in order to use it."
            )

        columns = list(model['dtype_dict'].keys())

        # cast where_data column to case of original predicto column
        if len(where_data) > 0:
            row = where_data[0]
            col_name_map = {}
            for i, col_name in enumerate(row):
                if col_name in ('__mindsdb_row_id', '__mdb_forecast_offset'):
                    continue
                new_col_name = get_column_in_case(columns, col_name)
                if new_col_name is not None and col_name != new_col_name:
                    col_name_map[col_name] = new_col_name
            if len(col_name_map) > 0:
                for row in where_data:
                    for old_col_name, new_col_name in col_name_map.items():
                        row[new_col_name] = row[old_col_name]
                        del row[old_col_name]

        predicted_columns = model['predict']
        if not isinstance(predicted_columns, list):
            predicted_columns = [predicted_columns]

        original_target_values = {}
        for col in predicted_columns:
            if where_data is not None:
                if col in where_data:
                    original_target_values[col + '_original'] = list(where_data[col])
                else:
                    original_target_values[col + '_original'] = [None] * len(where_data)
            else:
                original_target_values[col + '_original'] = [None]

        pred_dicts, explanations = self.model_interface.predict(table, where_data, 'dict&explain')

        # transform predictions to more convenient view
        new_pred_dicts = []
        for row in pred_dicts:
            new_row = {}
            for key in row:
                new_row.update(row[key])
                new_row[key] = new_row['predicted_value']
            del new_row['predicted_value']
            new_pred_dicts.append(new_row)
        pred_dicts = new_pred_dicts

        timeseries_settings = model['problem_definition']['timeseries_settings']

        if timeseries_settings['is_timeseries'] is True:
            __no_forecast_offset = set([row.get('__mdb_forecast_offset', None) for row in where_data]) == {None}

            predict = model['predict']
            group_by = timeseries_settings['group_by'] or []
            order_by_column = timeseries_settings['order_by'][0]
            horizon = timeseries_settings['horizon']

            groups = set()
            for row in pred_dicts:
                groups.add(
                    tuple([row[x] for x in group_by])
                )

            # split rows by groups
            rows_by_groups = {}
            for group in groups:
                rows_by_groups[group] = {
                    'rows': [],
                    'explanations': []
                }
                for row_index, row in enumerate(pred_dicts):
                    is_wrong_group = False
                    for i, group_by_key in enumerate(group_by):
                        if row[group_by_key] != group[i]:
                            is_wrong_group = True
                            break
                    if not is_wrong_group:
                        rows_by_groups[group]['rows'].append(row)
                        rows_by_groups[group]['explanations'].append(explanations[row_index])

            for group, data in rows_by_groups.items():
                rows = data['rows']
                explanations = data['explanations']

                if len(rows) == 0:
                    break

                for row in rows:
                    predictions = row[predict]
                    if isinstance(predictions, list) is False:
                        predictions = [predictions]

                    date_values = row[order_by_column]
                    if isinstance(date_values, list) is False:
                        date_values = [date_values]

                for i in range(len(rows) - 1):
                    if horizon > 1:
                        rows[i][predict] = rows[i][predict][0]
                        rows[i][order_by_column] = rows[i][order_by_column][0]
                    for col in ('predicted_value', 'confidence', 'confidence_lower_bound', 'confidence_upper_bound'):
                        if horizon > 1:
                            explanations[i][predict][col] = explanations[i][predict][col][0]

                last_row = rows.pop()
                last_explanation = explanations.pop()
                for i in range(horizon):
                    new_row = copy.deepcopy(last_row)
                    if horizon > 1:
                        new_row[predict] = new_row[predict][i]
                        new_row[order_by_column] = new_row[order_by_column][i]
                    if '__mindsdb_row_id' in new_row and (i > 0 or __no_forecast_offset):
                        new_row['__mindsdb_row_id'] = None
                    rows.append(new_row)

                    new_explanation = copy.deepcopy(last_explanation)
                    for col in ('predicted_value', 'confidence', 'confidence_lower_bound', 'confidence_upper_bound'):
                        if horizon > 1:
                            new_explanation[predict][col] = new_explanation[predict][col][i]
                    if i != 0:
                        new_explanation[predict]['anomaly'] = None
                        new_explanation[predict]['truth'] = None
                    explanations.append(new_explanation)

            pred_dicts = []
            explanations = []
            for group, data in rows_by_groups.items():
                pred_dicts.extend(data['rows'])
                explanations.extend(data['explanations'])

            original_target_values[f'{predict}_original'] = []
            for i in range(len(pred_dicts)):
                original_target_values[f'{predict}_original'].append(explanations[i][predict].get('truth', None))

            if model['dtypes'][order_by_column] == dtype.date:
                for row in pred_dicts:
                    if isinstance(row[order_by_column], (int, float)):
                        row[order_by_column] = str(datetime.fromtimestamp(row[order_by_column]).date())
            elif model['dtypes'][order_by_column] == dtype.datetime:
                for row in pred_dicts:
                    if isinstance(row[order_by_column], (int, float)):
                        row[order_by_column] = str(datetime.fromtimestamp(row[order_by_column]))

        keys = [x for x in pred_dicts[0] if x in columns]
        min_max_keys = []
        for col in predicted_columns:
            if model['dtype_dict'][col] in (dtype.integer, dtype.float, dtype.num_tsarray):
                min_max_keys.append(col)

        data = []
        explains = []
        keys_to_save = [*keys, '__mindsdb_row_id', 'select_data_query', 'when_data']
        for i, el in enumerate(pred_dicts):
            data.append({key: el.get(key) for key in keys_to_save})
            explains.append(explanations[i])

        for i, row in enumerate(data):
            cast_row_types(row, model['dtype_dict'])

            row['select_data_query'] = select_data_query
            row['when_data'] = original_when_data

            for k in original_target_values:
                try:
                    row[k] = original_target_values[k][i]
                except Exception:
                    row[k] = None

            for column_name in columns:
                if column_name not in row:
                    row[column_name] = None

            explanation = explains[i]
            for key in predicted_columns:
                row[key + '_confidence'] = explanation[key]['confidence']
                row[key + '_explain'] = json.dumps(explanation[key], cls=NumpyJSONEncoder, ensure_ascii=False)
                if 'anomaly' in explanation[key]:
                    row[key + '_anomaly'] = explanation[key]['anomaly']
            for key in min_max_keys:
                if 'confidence_lower_bound' in explanation[key]:
                    row[key + '_min'] = explanation[key]['confidence_lower_bound']
                if 'confidence_upper_bound' in explanation[key]:
                    row[key + '_max'] = explanation[key]['confidence_upper_bound']

        return data



import atexit
import traceback
import sys
import os
import time
import asyncio
import signal
import psutil
import json
import shutil

import torch.multiprocessing as mp
mp.set_start_method('spawn')
from packaging import version

from mindsdb.api.http.start import start as start_http
from mindsdb.api.mysql.start import start as start_mysql
from mindsdb.api.mongo.start import start as start_mongo
from mindsdb.utilities.config import Config
from mindsdb.utilities.ps import is_pid_listen_port, get_child_pids
from mindsdb.utilities.functions import args_parse, get_versions_where_predictors_become_obsolete
from mindsdb.utilities.with_kwargs_wrapper import WithKWArgsWrapper
from mindsdb.utilities.log import log
from mindsdb.interfaces.stream.stream import StreamController
from mindsdb.interfaces.stream.utilities import STOP_THREADS_EVENT
from mindsdb.interfaces.model.model_interface import ray_based, ModelInterface
from mindsdb.interfaces.database.integrations import IntegrationController
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.utilities.install import install_dependencies


COMPANY_ID = os.environ.get('MINDSDB_COMPANY_ID', None)


def close_api_gracefully(apis):
    try:
        for api in apis.values():
            process = api['process']
            childs = get_child_pids(process.pid)
            for p in childs:
                try:
                    os.kill(p, signal.SIGTERM)
                except Exception:
                    p.kill()
            sys.stdout.flush()
            process.terminate()
            process.join()
            sys.stdout.flush()
        if ray_based:
            os.system('ray stop --force')
    except KeyboardInterrupt:
        sys.exit(0)
    except psutil.NoSuchProcess:
        pass


if __name__ == '__main__':
    mp.freeze_support()
    args = args_parse()
    config = Config()

    is_cloud = config.get('cloud', False)

    if not is_cloud:
        print('Applying database migrations:')
        try:
            from mindsdb.migrations import migrate
            migrate.migrate_to_head()
        except Exception as e:
            print(f'Error! Something went wrong during DB migrations: {e}')

    if args.verbose is True:
        # Figure this one out later
        pass

    integration_controller = WithKWArgsWrapper(IntegrationController(), company_id=COMPANY_ID)
    if args.install_handlers is not None:
        handlers_list = [s.strip() for s in args.install_handlers.split(',')]
        # import_meta = handler_meta.get('import', {})
        for handler_name, handler_meta in integration_controller.get_handlers_import_status().items():
            if handler_name not in handlers_list:
                continue
            import_meta = handler_meta.get('import', {})
            if import_meta.get('success') is True:
                print(f"{'{0: <18}'.format(handler_name)} - already installed")
                continue
            result = install_dependencies(import_meta.get('dependencies', []))
            if result.get('success') is True:
                print(f"{'{0: <18}'.format(handler_name)} - successfully installed")
            else:
                print(f"{'{0: <18}'.format(handler_name)} - error during dependencies installation: {result.get('error_message', 'unknown error')}")
        sys.exit(0)

    os.environ['DEFAULT_LOG_LEVEL'] = config['log']['level']['console']
    os.environ['LIGHTWOOD_LOG_LEVEL'] = config['log']['level']['console']

    # Switch to this once the native interface has it's own thread :/
    ctx = mp.get_context('spawn')

    from mindsdb.__about__ import __version__ as mindsdb_version
    print(f'Version {mindsdb_version}')

    print(f'Configuration file:\n   {config.config_path}')
    print(f"Storage path:\n   {config['paths']['root']}")

    # @TODO Backwards compatibiltiy for tests, remove later
    model_interface = WithKWArgsWrapper(ModelInterface(), company_id=COMPANY_ID)
    integration_controller = WithKWArgsWrapper(IntegrationController(), company_id=COMPANY_ID)
    for handler_name, handler_meta in integration_controller.get_handlers_import_status().items():
        import_meta = handler_meta.get('import', {})
        if import_meta.get('success', False) is not True:
            print(f"Can't import handler '{handler_name}': {import_meta.get('error_message', 'unknown error')}")

    raw_model_data_arr = model_interface.get_models()
    model_data_arr = []
    for model in raw_model_data_arr:
        if model['status'] == 'complete':
            x = model_interface.get_model_data(model['name'])
            try:
                model_data_arr.append(model_interface.get_model_data(model['name']))
            except Exception:
                pass

    if not is_cloud:
        # region creating permanent integrations
        for integration_name in ['files', 'views']:
            integration_meta = integration_controller.get(name=integration_name)
            if integration_meta is None:
                integration_record = db.Integration(
                    name=integration_name,
                    data={},
                    engine=integration_name,
                    company_id=None
                )
                db.session.add(integration_record)
                db.session.commit()
        # endregion

        # region Mark old predictors as outdated
        is_modified = False
        predictor_records = db.session.query(db.Predictor).all()
        if len(predictor_records) > 0:
            sucess, compatible_versions = get_versions_where_predictors_become_obsolete()
            if sucess is True:
                compatible_versions = [version.parse(x) for x in compatible_versions]
                mindsdb_version_parsed = version.parse(mindsdb_version)
                compatible_versions = [x for x in compatible_versions if x <= mindsdb_version_parsed]
                if len(compatible_versions) > 0:
                    last_compatible_version = compatible_versions[-1]
                    for predictor_record in predictor_records:
                        if (
                            isinstance(predictor_record.mindsdb_version, str) is not None
                            and version.parse(predictor_record.mindsdb_version) < last_compatible_version
                        ):
                            predictor_record.update_status = 'available'
                            is_modified = True
        if is_modified is True:
            db.session.commit()
        # endregion

        for integration_name in config.get('integrations', {}):
            try:
                it = integration_controller.get(integration_name)
                if it is not None:
                    integration_controller.delete(integration_name)
                print(f'Adding: {integration_name}')
                integration_data = config['integrations'][integration_name]
                engine = integration_data.get('type')
                if engine is not None:
                    del integration_data['type']
                integration_controller.add(integration_name, engine, integration_data)
            except Exception as e:
                log.error(f'\n\nError: {e} adding database integration {integration_name}\n\n')

        stream_controller = StreamController(COMPANY_ID)
        for integration_name, integration_meta in integration_controller.get_all(sensitive_info=True).items():
            if (
                integration_meta.get('type') in stream_controller.known_dbs
                and integration_meta.get('publish', False) is True
            ):
                print(f"Setting up stream: {integration_name}")
                stream_controller.setup(integration_name)
        del stream_controller

    del model_interface
    # @TODO Backwards compatibiltiy for tests, remove later

    if args.api is None:
        api_arr = ['http', 'mysql']
    else:
        api_arr = args.api.split(',')

    with_nlp = False
    if 'nlp' in api_arr:
        with_nlp = True
        api_arr.remove('nlp')

    apis = {
        api: {
            'port': config['api'][api]['port'],
            'process': None,
            'started': False
        } for api in api_arr
    }

    start_functions = {
        'http': start_http,
        'mysql': start_mysql,
        'mongodb': start_mongo
    }

    for api_name, api_data in apis.items():
        if api_data['started']:
            continue
        print(f'{api_name} API: starting...')
        try:
            if api_name == 'http':
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose, args.no_studio, with_nlp))
            else:
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose,))
            p.start()
            api_data['process'] = p
        except Exception as e:
            log.error(f'Failed to start {api_name} API with exception {e}\n{traceback.format_exc()}')
            close_api_gracefully(apis)
            raise e

    atexit.register(close_api_gracefully, apis=apis)

    async def wait_api_start(api_name, pid, port):
        timeout = 60
        start_time = time.time()
        started = is_pid_listen_port(pid, port)
        while (time.time() - start_time) < timeout and started is False:
            await asyncio.sleep(0.5)
            started = is_pid_listen_port(pid, port)
        return api_name, port, started

    async def wait_apis_start():
        futures = [
            wait_api_start(api_name, api_data['process'].pid, api_data['port'])
            for api_name, api_data in apis.items() if 'port' in api_data
        ]
        for i, future in enumerate(asyncio.as_completed(futures)):
            api_name, port, started = await future
            if started:
                print(f"{api_name} API: started on {port}")
            else:
                log.error(f"ERROR: {api_name} API cant start on {port}")

    ioloop = asyncio.get_event_loop()
    ioloop.run_until_complete(wait_apis_start())
    ioloop.close()

    try:
        for api_data in apis.values():
            api_data['process'].join()
    except KeyboardInterrupt:
        print('Stopping stream integrations...')
        STOP_THREADS_EVENT.set()
        print('Closing app...')



"""predictor-integration

Revision ID: d74c189b87e6
Revises: 27c5aca9e47e
Create Date: 2022-05-25 15:00:16.284158

"""
import json

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text


# revision identifiers, used by Alembic.
revision = 'd74c189b87e6'
down_revision = '27c5aca9e47e'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('integration_id', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('fetch_data_query', sa.String(), nullable=True))
        batch_op.create_foreign_key('fk_integration_id', 'integration', ['integration_id'], ['id'])

    conn = op.get_bind()

    conn.execute(text('''
        insert into integration (name, data, company_id, created_at, updated_at)
            select
                'files' as name,
                '{}' as data,
                company_id,
                '2022-05-01 00:00:00.000000' as created_at,
                '2022-05-01 00:00:00.000000' as updated_at
            from (select distinct company_id from integration) t1
    '''))

    predictors = conn.execute(text('''
        select t1.id, t1.company_id, t2.data
        from predictor t1 left join dataset t2 on t1.id = t2.id
        where dataset_id is not null and t2.data is not null
    ''')).fetchall()
    for row in predictors:
        data = row['data']
        try:
            data = json.loads(data)
        except Exception:
            continue

        if 'source_type' not in data:
            continue
        integration_name = data.get('source_type')
        if isinstance(integration_name, str) is False or len(integration_name) == 0:
            continue
        if integration_name.lower() == 'file':
            integration_name = 'files'

        fetch_data_query = data.get('source')
        if isinstance(fetch_data_query, dict) is False:
            continue

        if integration_name == 'files':
            file_name = fetch_data_query.get('mindsdb_file_name')
            if isinstance(file_name, str) is False or len(file_name) == 0:
                continue
            fetch_data_query = f'select * from {file_name}'
        else:
            fetch_data_query = fetch_data_query.get('query')
            if isinstance(fetch_data_query, str) is False or len(fetch_data_query) == 0:
                continue

        query = '''
            select id
            from integration
            where company_id = :company_id and lower(name) = lower(:name)
        '''
        if row['company_id'] is None:
            query = '''
                select id
                from integration
                where company_id is null and lower(name) = lower(:name)
            '''
        integration = conn.execute(text(query), {
            'company_id': row['company_id'],
            'name': integration_name
        }).fetchone()
        if integration is None:
            continue

        conn.execute(text('''
            update predictor
            set integration_id = :integration_id, fetch_data_query = :fetch_data_query
            where id = :predictor_id
        '''), {
            'integration_id': integration.id,
            'fetch_data_query': fetch_data_query,
            'predictor_id': row['id']
        })

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('dataset_id')
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.drop_constraint('unique_file_name_company_id', type_='unique')
        batch_op.drop_constraint('fk_analysis_id', type_='foreignkey')
        batch_op.drop_column('analysis_id')
    op.drop_table('analysis')
    op.drop_table('dataset')


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('dataset_id', sa.INTEGER(), nullable=True))
        batch_op.drop_constraint('fk_integration_id', type_='foreignkey')
        batch_op.create_foreign_key('fk_predictor_dataset_id', 'dataset', ['dataset_id'], ['id'])
        batch_op.drop_column('fetch_data_query')
        batch_op.drop_column('integration_id')

    op.create_table(
        'analysis',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('analysis', sa.VARCHAR(), nullable=False),
        sa.Column('created_at', sa.DATETIME(), nullable=True),
        sa.Column('updated_at', sa.DATETIME(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'dataset',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('updated_at', sa.DATETIME(), nullable=True),
        sa.Column('created_at', sa.DATETIME(), nullable=True),
        sa.Column('name', sa.VARCHAR(), nullable=True),
        sa.Column('data', sa.VARCHAR(), nullable=True),
        sa.Column('creation_info', sa.VARCHAR(), nullable=True),
        sa.Column('company_id', sa.INTEGER(), nullable=True),
        sa.Column('mindsdb_version', sa.VARCHAR(), nullable=True),
        sa.Column('datasources_version', sa.VARCHAR(), nullable=True),
        sa.Column('integration_id', sa.INTEGER(), nullable=True),
        sa.Column('analysis_id', sa.INTEGER(), nullable=True),
        sa.Column('ds_class', sa.VARCHAR(), nullable=True),
        sa.ForeignKeyConstraint(['analysis_id'], ['analysis.id'], name='fk_analysis_id'),
        sa.ForeignKeyConstraint(['integration_id'], ['integration.id'], name='fk_integration_id'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_dataset_name_company_id')
    )

    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.add_column(sa.Column('analysis_id', sa.INTEGER(), nullable=True))
        batch_op.create_foreign_key('fk_analysis_id', 'analysis', ['analysis_id'], ['id'])
        batch_op.create_unique_constraint('unique_file_name_company_id', ['name', 'company_id'])



import shutil
import os
from mindsdb.utilities.config import Config
from checksumdir import dirhash
import hashlib


def copy(src, dst):
    if os.path.isdir(src):
        if os.path.exists(dst):
            if dirhash(src) == dirhash(dst):
                return
        shutil.rmtree(dst, ignore_errors=True)
        shutil.copytree(src, dst)
    else:
        if os.path.exists(dst):
            if hashlib.md5(open(src, 'rb').read()).hexdigest() == hashlib.md5(open(dst, 'rb').read()).hexdigest():
                return
        try:
            os.remove(dst)
        except Exception:
            pass
        shutil.copy2(src, dst)


try:
    import boto3
except Exception:
    # Only required for remote storage on s3
    pass


class FsStore():
    def __init__(self):
        self.config = Config()
        self.location = self.config['permanent_storage']['location']
        if self.location == 'local':
            pass
        elif self.location == 's3':
            if 's3_credentials' in self.config['permanent_storage']:
                self.s3 = boto3.client('s3', **self.config['permanent_storage']['s3_credentials'])
            else:
                self.s3 = boto3.client('s3')
            self.bucket = self.config['permanent_storage']['bucket']
        else:
            raise Exception('Location: ' + self.location + ' not supported')

    def put(self, filename, remote_name, local_path):
        if self.location == 'local':
            print('From: ', os.path.join(local_path, filename))
            print('To: ', os.path.join(self.config['paths']['storage'], remote_name))
            copy(os.path.join(local_path, filename), os.path.join(self.config['paths']['storage'], remote_name))
        elif self.location == 's3':
            # NOTE: This `make_archive` function is implemente poorly and will create an empty archive file even if the file/dir to be archived doesn't exist or for some other reason can't be archived
            shutil.make_archive(os.path.join(local_path, remote_name), 'gztar', root_dir=local_path, base_dir=filename)
            self.s3.upload_file(os.path.join(local_path, f'{remote_name}.tar.gz'), self.bucket, f'{remote_name}.tar.gz')
            os.remove(os.path.join(local_path, remote_name + '.tar.gz'))

    def get(self, filename, remote_name, local_path):
        if self.location == 'local':
            copy(os.path.join(self.config['paths']['storage'], remote_name), os.path.join(local_path, filename))
        elif self.location == 's3':
            remote_ziped_name = f'{remote_name}.tar.gz'
            local_ziped_name = f'{filename}.tar.gz'
            local_ziped_path = os.path.join(local_path, local_ziped_name)
            self.s3.download_file(self.bucket, remote_ziped_name, local_ziped_path)
            shutil.unpack_archive(local_ziped_path, local_path)
            os.system(f'chmod -R 777 {local_path}')
            os.remove(local_ziped_path)

    def delete(self, remote_name):
        if self.location == 'local':
            pass
        elif self.location == 's3':
            self.s3.delete_object(Bucket=self.bucket, Key=remote_name)



from bson.int64 import Int64

from mindsdb_sql.parser.ast import *
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes import Responder
from mindsdb.integrations.handlers.mongodb_handler.utils.mongodb_ast import MongoToAst

from mindsdb.api.mongo.classes.query_sql import run_sql_command


class Responce(Responder):
    when = {'find': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        database = request_env['database']

        # system queries
        if query['find'] == 'system.version':
            # For studio3t
            data = [{
                "_id": "featureCompatibilityVersion",
                "version": "3.6"
            }]
            cursor = {
                'id': Int64(0),
                'ns': f"system.version.$cmd.{query['find']}",
                'firstBatch': data
            }
            return {
                'cursor': cursor,
                'ok': 1
            }

        mongoToAst = MongoToAst()

        collection = [database, query['find']]

        if not query.get('singleBatch') and 'collection' in query.get('filter'):
            # JOIN mode

            # upper query
            ast_query = mongoToAst.find(
                collection=collection,
                projection=query.get('projection'),
                sort=query.get('sort'),
                limit=query.get('limit'),
                skip=query.get('skip'),
            )

            # table_query
            collection = query['filter']['collection']
            filter = query['filter']['query']
            table_select = mongoToAst.find(
                collection=collection,
                filter=filter,
            )
            table_select.parentheses = True

            # convert to join
            right_table = ast_query.from_table

            ast_join = Join(
                left=table_select,
                right=right_table,
                join_type='join'
            )
            ast_query.from_table = ast_join

        else:
            # is single table
            ast_query = mongoToAst.find(
                collection=collection,
                filter=query.get('filter'),
                projection=query.get('projection'),
                sort=query.get('sort'),
                limit=query.get('limit'),
                skip=query.get('skip'),
            )

        data = run_sql_command(mindsdb_env, ast_query)

        db = mindsdb_env['config']['api']['mongodb']['database']

        cursor = {
            'id': Int64(0),
            'ns': f"{db}.$cmd.{query['find']}",
            'firstBatch': data
        }
        return {
            'cursor': cursor,
            'ok': 1
        }


responder = Responce()



import base64
from datetime import datetime, date, timedelta

import numpy as np
from flask.json import JSONEncoder


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%dT%H:%M:%S.%f")
        if isinstance(obj, timedelta):
            return str(obj)
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.int8) or isinstance(obj, np.int16) or isinstance(obj, np.int32) or isinstance(obj, np.int64):
            return int(obj)
        if isinstance(obj, np.float16) or isinstance(obj, np.float32) or isinstance(obj, np.float64) or isinstance(obj, np.float128):
            return float(obj)

        return JSONEncoder.default(self, obj)


def json_serialiser(byte_obj):
    """
    Used to export/import predictors inside the model controller.
    Reference: https://stackoverflow.com/q/53942948.
    """
    if isinstance(byte_obj, (bytes, bytearray)):
        # File Bytes to Base64 Bytes then to String
        return base64.b64encode(byte_obj).decode('utf-8')
    raise ValueError('No encoding handler for data type ' + type(byte_obj))



from .responder import Responder


class RespondersCollection():
    def __init__(self):
        self.responders = []

    def find_match(self, query):
        for r in self.responders:
            if r.match(query):
                return r
        raise Exception(f'Is not responder for query: {query}')

    def add(self, when, result):
        self.responders.append(
            Responder(when, result)
        )



import copy
from itertools import product
import pandas as pd

from .ts_utils import validate_ts_where_condition, find_time_filter, add_order_not_null, replace_time_filter, find_and_remove_time_filter, get_time_selects

from mindsdb_sql.parser.ast import Identifier, Constant, Operation, Select, BinaryOperation, BetweenOperation
from mindsdb_sql.parser.ast import OrderBy


def get_ts_join_input(query, model, data_handler, data_side):
    # TODO: TS query coverage should equal pre-handler release

    # step 1) query checks
    if query.order_by:
        raise PlanningException(
            f'Can\'t provide ORDER BY to time series predictor join query. Found: {query.order_by}.')

    if query.group_by or query.having or query.offset:
        raise PlanningException(f'Unsupported query to timeseries predictor: {str(query)}')

    if not model.problem_definition.timeseries_settings.is_timeseries:
        raise PlanningException(f"This is not a time-series predictor, aborting.")

    data_handler_table = getattr(query.from_table, data_side).parts[-1]
    data_handler_alias = getattr(query.from_table, data_side).alias
    data_handler_cols = list(set([t.parts[-1] for t in query.targets if t.parts[0] == str(data_handler_alias)]))

    window = model.problem_definition.timeseries_settings.window
    oby_col = model.problem_definition.timeseries_settings.order_by[0]
    gby_cols = model.problem_definition.timeseries_settings.group_by

    allowed_columns = [oby_col.lower()]
    if len(gby_cols) > 0:
        allowed_columns += [i.lower() for i in gby_cols]
    validate_ts_where_condition(query.where, allowed_columns=allowed_columns)

    time_filter = find_time_filter(query.where, time_column_name=oby_col)
    order_by = [OrderBy(Identifier(parts=[oby_col]), direction='DESC')]

    # step 2) get time filter
    preparation_where = copy.deepcopy(query.where)
    preparation_where = add_order_not_null(preparation_where, time_column_name=oby_col)
    time_selects = get_time_selects(time_filter, data_handler_table, window, order_by, preparation_where)

    # step 3) execute time filter on all required partitions
    if len(gby_cols) == 0:
        # no groups - one or multistep
        if len(time_selects) == 1:
            model_input = pd.DataFrame.from_records(data_handler.query(time_selects[0]).data_frame)
        else:
            dfs = []
            for step in time_selects:
                # TODO: correctness if we have a double cutoff?
                dfs.append(pd.DataFrame.from_records(data_handler.query(step).data_frame))
            model_input = pd.concat(dfs)
    else:
        # grouped - multiple partitions
        groups = {}
        dfs = []
        for gcol in gby_cols:
            groups_query = Select(
                targets=[Identifier(gcol)],
                distinct=True,
                from_table=Identifier(data_handler_table),
            )
            groups[gcol] = list(data_handler.query(groups_query).data_frame.squeeze().values)

        partition_keys = list(groups.keys())
        all_partitions = list(product(*[v for k, v in groups.items()]))  # TODO: check for better retrival then project?

        for group in all_partitions:
            group_time_selects = copy.deepcopy(time_selects)

            # TODO: pending
            # # one or multistep
            # if len(group_time_selects) == 1:
            #     partial_df = partial_dfs[0]
            # else:
            #     partial_df = pd.concat(partial_dfs)
            #
            # # get grouping values
            # # TODO: keep time filter removal?
            # no_time_filter_query = copy.deepcopy(query)
            # no_time_filter_query.where = find_and_remove_time_filter(no_time_filter_query.where, time_filter)

            filters = None
            for i, val in enumerate(group):
                col = partition_keys[i]
                binop = BinaryOperation(op='=',
                                        args=[
                                            Identifier(col),
                                            Constant(val)
                                        ])
                if filters is None:
                    filters = binop
                else:
                    filters = BinaryOperation(op='and', args=[filters, binop])

            for time_select in group_time_selects:
                # TODO: this doesn't cover intersection case...
                time_select.where = BinaryOperation(op='and', args=[time_select.where, filters])

                df = data_handler.query(time_select).data_frame
                # TODO: maybe order and limit df instead of SELECT to hedge against badly defined dtypes in the DB?
                dfs.append(df)

        # 3) concatenate all contexts into single data query
        model_input = pd.concat(dfs).reset_index(drop=True)

    return model_input


"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import re
from collections import OrderedDict, defaultdict
import datetime
import time

import duckdb
import pandas as pd
import numpy as np
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import (
    BinaryOperation,
    UnaryOperation,
    CreateTable,
    Identifier,
    Constant,
    Select,
    Union,
    Join,
    Star,
    Insert,
    Delete,
)
from mindsdb_sql.planner.steps import (
    ApplyTimeseriesPredictorStep,
    ApplyPredictorRowStep,
    GetPredictorColumns,
    FetchDataframeStep,
    ApplyPredictorStep,
    GetTableColumns,
    LimitOffsetStep,
    MapReduceStep,
    MultipleSteps,
    ProjectStep,
    SaveToTable,
    InsertToTable,
    FilterStep,
    UnionStep,
    JoinStep,
    GroupByStep
)

from mindsdb_sql.exceptions import PlanningException
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.planner import query_planner
from mindsdb_sql.planner.utils import query_traversal

from mindsdb.api.mysql.mysql_proxy.classes.com_operators import operator_map
import mindsdb.interfaces.storage.db as db
from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.api.mysql.mysql_proxy.utilities.functions import get_column_in_case

from mindsdb.api.mysql.mysql_proxy.utilities import (
    SqlApiException,
    ErKeyColumnDoesNotExist,
    ErNotSupportedYet,
    SqlApiUnknownError,
    ErLogicError,
    ErSqlWrongArguments
)
from mindsdb_sql.parser.ast.base import ASTNode


superset_subquery = re.compile(r'from[\s\n]*(\(.*\))[\s\n]*as[\s\n]*virtual_table', flags=re.IGNORECASE | re.MULTILINE | re.S)


def get_preditor_alias(step, mindsdb_database):
    predictor_name = '.'.join(step.predictor.parts)
    predictor_alias = '.'.join(step.predictor.alias.parts) if step.predictor.alias is not None else predictor_name
    return (mindsdb_database, predictor_name, predictor_alias)


def get_table_alias(table_obj, default_db_name):
    # (database, table, alias)
    if isinstance(table_obj, Identifier):
        if len(table_obj.parts) > 2:
            raise ErSqlWrongArguments(f'Table name must contain no more than 2 parts. Got name: {table_obj.parts}')
        elif len(table_obj.parts) == 1:
            name = (default_db_name, table_obj.parts[0])
        else:
            name = tuple(table_obj.parts)
    else:
        # it is subquery
        name = table_obj.alias.parts[0] or 't'
        name = (default_db_name, name)

    if table_obj.alias is not None:
        name = name + ('.'.join(table_obj.alias.parts),)
    else:
        name = name + (name[1],)
    return name


def get_all_tables(stmt):
    if isinstance(stmt, Union):
        left = get_all_tables(stmt.left)
        right = get_all_tables(stmt.right)
        return left + right

    if isinstance(stmt, Select):
        from_stmt = stmt.from_table
    elif isinstance(stmt, (Identifier, Join)):
        from_stmt = stmt
    elif isinstance(stmt, Insert):
        from_stmt = stmt.table
    elif isinstance(stmt, Delete):
        from_stmt = stmt.table
    else:
        # raise SqlApiException(f'Unknown type of identifier: {stmt}')
        return []

    result = []
    if isinstance(from_stmt, Identifier):
        result.append(from_stmt.parts[-1])
    elif isinstance(from_stmt, Join):
        result.extend(get_all_tables(from_stmt.left))
        result.extend(get_all_tables(from_stmt.right))
    return result


def markQueryVar(where):
    if isinstance(where, BinaryOperation):
        markQueryVar(where.args[0])
        markQueryVar(where.args[1])
    elif isinstance(where, UnaryOperation):
        markQueryVar(where.args[0])
    elif isinstance(where, Constant):
        if str(where.value).startswith('$var['):
            where.is_var = True
            where.var_name = where.value


def unmarkQueryVar(where):
    if isinstance(where, BinaryOperation):
        unmarkQueryVar(where.args[0])
        unmarkQueryVar(where.args[1])
    elif isinstance(where, UnaryOperation):
        unmarkQueryVar(where.args[0])
    elif isinstance(where, Constant):
        if hasattr(where, 'is_var') and where.is_var is True:
            where.value = where.var_name


def replaceQueryVar(where, var_value, var_name):
    if isinstance(where, BinaryOperation):
        replaceQueryVar(where.args[0], var_value, var_name)
        replaceQueryVar(where.args[1], var_value, var_name)
    elif isinstance(where, UnaryOperation):
        replaceQueryVar(where.args[0], var_value, var_name)
    elif isinstance(where, Constant):
        if hasattr(where, 'is_var') and where.is_var is True and where.value == f'$var[{var_name}]':
            where.value = var_value


def join_query_data(target, source):
    target['values'].extend(source['values'])
    target['tables'].extend(source['tables'])
    target['tables'] = list(set(target['tables']))
    for table_name in source['columns']:
        if table_name not in target['columns']:
            target['columns'][table_name] = source['columns'][table_name]
        else:
            target['columns'][table_name].extend(source['columns'][table_name])
            target['columns'][table_name] = list(set(target['columns'][table_name]))


def is_empty_prediction_row(predictor_value):
    "Define empty rows in predictor after JOIN"
    for key in predictor_value:
        if predictor_value[key] is not None and pd.notna(predictor_value[key]):
            return False
    return True


class Column:
    def __init__(self, name=None, alias=None,
                 table_name=None, table_alias=None,
                 type=None, database=None, flags=None,
                 charset=None):
        if alias is None:
            alias = name
        if table_alias is None:
            table_alias = table_name
        self.name = name
        self.alias = alias
        self.table_name = table_name
        self.table_alias = table_alias
        self.type = type
        self.database = database
        self.flags = flags
        self.charset = charset

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__})'


class SQLQuery():
    def __init__(self, sql, session, execute=True):
        self.session = session
        self.integration = session.integration
        self.database = None if session.database == '' else session.database.lower()
        self.datahub = session.datahub
        self.outer_query = None
        self.row_id = 0
        self.columns_list = None
        self.model_types = {}

        self.mindsdb_database_name = 'mindsdb'

        if isinstance(sql, str):
        # +++ workaround for subqueries in superset
            if 'as virtual_table' in sql.lower():
                subquery = re.findall(superset_subquery, sql)
                if isinstance(subquery, list) and len(subquery) == 1:
                    subquery = subquery[0]
                    self.outer_query = sql.replace(subquery, 'dataframe')
                    sql = subquery.strip('()')
            # ---
            self.query = parse_sql(sql, dialect='mindsdb')
            self.query_str = sql
        else:
            self.query = sql
            renderer = SqlalchemyRender('mysql')
            try:
                self.query_str = renderer.get_string(self.query, with_failback=True)
            except Exception:
                self.query_str = str(self.query)

        self.planner = None
        self.parameters = []
        self.fetched_data = None
        # self._process_query(sql)
        self.create_planner()
        if execute:
            self.prepare_query(prepare=False)
            self.execute_query()

    def create_planner(self):

        integrations_names = self.session.datahub.get_integrations_names()
        integrations_names.append('information_schema')
        integrations_names.append('files')
        integrations_names.append('views')

        predictor_metadata = {}
        predictors = db.session.query(db.Predictor).filter_by(company_id=self.session.company_id)

        query_tables = []

        def get_all_query_tables(node, is_table, **kwargs):
            if is_table and isinstance(node, Identifier):
                query_tables.append(node.parts[-1])

        query_traversal(self.query, get_all_query_tables)

        # get all predictors
        for p in predictors:
            model_name = p.name

            if model_name not in query_tables:
                # skip
                continue

            if isinstance(p.data, dict) and 'error' not in p.data:
                ts_settings = p.learn_args.get('timeseries_settings', {})
                if ts_settings.get('is_timeseries') is True:
                    window = ts_settings.get('window')
                    order_by = ts_settings.get('order_by')[0]
                    group_by = ts_settings.get('group_by')
                    if isinstance(group_by, list) is False and group_by is not None:
                        group_by = [group_by]
                    predictor_metadata[model_name] = {
                        'timeseries': True,
                        'window': window,
                        'horizon': ts_settings.get('horizon'),
                        'order_by_column': order_by,
                        'group_by_columns': group_by
                    }
                else:
                    predictor_metadata[model_name] = {
                        'timeseries': False
                    }
                self.model_types.update(p.data.get('dtypes', {}))

        mindsdb_database_name = 'mindsdb'
        database = None if self.session.database == '' else self.session.database.lower()

        self.planner = query_planner.QueryPlanner(
            self.query,
            integrations=integrations_names,
            predictor_namespace=mindsdb_database_name,
            predictor_metadata=predictor_metadata,
            default_namespace=database
        )

    def fetch(self, datahub, view='list'):
        data = self.fetched_data

        if view == 'list':
            self.result = self._make_list_result_view(data)

        # this is not used
        # elif view == 'dict':
        #     self.result = self._make_dict_result_view(data)
        # else:
        #     raise ErNotSupportedYet('Only "list" and "dict" views supported atm')

        return {
            'success': True,
            'result': self.result
        }

    def _fetch_dataframe_step(self, step):
        dn = self.datahub.get(step.integration)
        query = step.query

        table_alias = get_table_alias(step.query.from_table, self.database)
        # TODO for information_schema we have 'database' = 'mindsdb'

        data, columns_info = dn.query(
            query=query
        )

        # if this is query: execute it
        if isinstance(data, ASTNode):
            subquery = SQLQuery(data, session=self.session)
            return subquery.fetched_data

        columns = [(column['name'], column['name']) for column in columns_info]
        columns.append(('__mindsdb_row_id', '__mindsdb_row_id'))

        for i, row in enumerate(data):
            row['__mindsdb_row_id'] = self.row_id + i
        self.row_id = self.row_id + len(data)

        data = [{(key, key): value for key, value in row.items()} for row in data]
        data = [{table_alias: x} for x in data]

        col_types = {
            column['name']: column['type']
            for column in columns_info
        }
        data = {
            'values': data,
            'columns': {table_alias: columns},
            'tables': [table_alias],
            'types': {table_alias: col_types}
        }
        return data

    def _multiple_steps(self, step):
        data = {
            'values': [],
            'columns': {},
            'tables': []
        }
        for substep in step.steps:
            sub_data = self._fetch_dataframe_step(substep)
            join_query_data(data, sub_data)
        return data

    def _multiple_steps_reduce(self, step, vars):
        if step.reduce != 'union':
            raise ErLogicError(f'Unknown MultipleSteps type: {step.reduce}')

        data = {
            'values': [],
            'columns': {},
            'tables': []
        }

        for var_group in vars:
            for substep in step.steps:
                if isinstance(substep, FetchDataframeStep) is False:
                    raise ErLogicError(f'Wrong step type for MultipleSteps: {step}')
                markQueryVar(substep.query.where)
            for name, value in var_group.items():
                for substep in step.steps:
                    replaceQueryVar(substep.query.where, value, name)
            sub_data = self._multiple_steps(step)
            join_query_data(data, sub_data)

        return data


    def prepare_query(self, prepare=True):
        mindsdb_sql_struct = self.query

        if isinstance(mindsdb_sql_struct, Select):
            # is it query to 'predictors'?
            if (
                isinstance(mindsdb_sql_struct.from_table, Identifier)
                and mindsdb_sql_struct.from_table.parts[-1].lower() == 'predictors'
                and (
                    self.database == 'mindsdb'
                    or mindsdb_sql_struct.from_table.parts[0].lower() == 'mindsdb'
                )
            ):
                dn = self.datahub.get(self.mindsdb_database_name)
                data, columns = dn.get_predictors(mindsdb_sql_struct)
                table_name = ('mindsdb', 'predictors', 'predictors')
                data = [
                    {
                        (key, key): value
                        for key, value in row.items()
                    }
                    for row in data
                ]
                data = [{table_name: x} for x in data]
                self.columns_list = [
                    Column(database='mindsdb',
                           table_name='predictors',
                           name=column_name)
                    for column_name in columns
                ]

                columns = [(column_name, column_name) for column_name in columns]

                self.fetched_data = {
                    'values': data,
                    'columns': {table_name: columns},
                    'tables': [table_name]
                }
                return

            # is it query to 'datasources'?
            if (
                isinstance(mindsdb_sql_struct.from_table, Identifier)
                and mindsdb_sql_struct.from_table.parts[-1].lower() in ('datasources', 'databases')
                and (
                    self.database == 'mindsdb'
                    or mindsdb_sql_struct.from_table.parts[0].lower() == 'mindsdb'
                )
            ):
                dn = self.datahub.get(self.mindsdb_database_name)
                data, columns = dn.get_integrations(mindsdb_sql_struct)
                table_name = ('mindsdb', 'datasources', 'datasources')
                data = [
                    {
                        (key, key): value
                        for key, value in row.items()
                    }
                    for row in data
                ]

                data = [{table_name: x} for x in data]

                self.columns_list = [
                    Column(database='mindsdb',
                           table_name='datasources',
                           name=column_name)
                    for column_name in columns
                ]

                columns = [(column_name, column_name) for column_name in columns]

                self.fetched_data = {
                    'values': data,
                    'columns': {table_name: columns},
                    'tables': [table_name]
                }
                return

        if prepare:
            # it is prepared statement call
            steps_data = []
            try:
                for step in self.planner.prepare_steps(self.query):
                    data = self.execute_step(step, steps_data)
                    step.set_result(data)
                    steps_data.append(data)
            except PlanningException as e:
                raise ErLogicError(e)

            statement_info = self.planner.get_statement_info()

            self.columns_list = []
            for col in statement_info['columns']:
                self.columns_list.append(
                    Column(
                        database=col['ds'],
                        table_name=col['table_name'],
                        table_alias=col['table_alias'],
                        name=col['name'],
                        alias=col['alias'],
                        type=col['type']
                    )
                )

            self.parameters = [
                Column(
                    name=col['name'],
                    alias=col['alias'],
                    type=col['type']
                )
                for col in statement_info['parameters']
            ]

    def execute_query(self, params=None):
        if self.fetched_data is not None:
            # no need to execute
            return

        steps_data = []
        try:
            for step in self.planner.execute_steps(params):
                data = self.execute_step(step, steps_data)
                step.set_result(data)
                steps_data.append(data)
        except PlanningException as e:
            raise ErLogicError(e)

        # save updated query
        self.query = self.planner.query

        # there was no executing
        if len(steps_data) == 0:
            return

        try:
            if self.outer_query is not None:
                data = []
                # +++
                # ???
                result = []
                for row in steps_data[-1]:
                    data_row = {}
                    for column_record in self.columns_list:
                        table_name = (column_record.database, column_record.table_name, column_record.table_alias)
                        column_name = column_record.name
                        data_row[column_record.alias or column_record.name] = row[table_name][column_name]
                    result.append(data_row)
                # ---
                data = self._make_list_result_view(result)
                df = pd.DataFrame(data)
                result = query_df(df, self.outer_query)

                try:
                    self.columns_list = [
                        Column(database='',
                               table_name='',
                               name=x)
                        for x in result.columns
                    ]
                except Exception:
                    self.columns_list = [
                        Column(database='',
                               table_name='',
                               name=result.name)
                    ]

                # +++ make list result view
                new_result = []
                for row in result.to_dict(orient='records'):
                    data_row = []
                    for column_record in self.columns_list:
                        column_name = column_record.alias or column_record.name
                        data_row.append(row.get(column_name))
                    new_result.append(data_row)
                result = new_result
                # ---

                self.fetched_data = result
            else:
                self.fetched_data = steps_data[-1]
        except Exception as e:
            raise SqlApiUnknownError("error in preparing result quiery step") from e

        try:
            if hasattr(self, 'columns_list') is False:
                self.columns_list = []
                keys = []
                for row in self.fetched_data:
                    for table_key in row:
                        for column_name in row[table_key]:
                            key = (table_key + (column_name, column_name))
                            if key not in keys:
                                keys.append(key)
                                self.columns_list.append(
                                    Column(database=table_key[0],
                                           table_name=table_key[1],
                                           table_alias=table_key[2],
                                           name=column_name)
                                )

            # if there was no 'ProjectStep', then get columns list from last step:
            if self.columns_list is None:
                self.columns_list = []
                if self.fetched_data is not None:
                    for table_name in self.fetched_data['columns']:
                        col_types = self.fetched_data.get('types', {}).get(table_name, {})
                        for column in self.fetched_data['columns'][table_name]:
                            self.columns_list.append(
                                Column(
                                    database=table_name[0],
                                    table_name=table_name[1],
                                    table_alias=table_name[2],
                                    name=column[0],
                                    alias=column[1],
                                    type=col_types.get(column[0])
                                )
                            )

            self.columns_list = [x for x in self.columns_list if x.name != '__mindsdb_row_id']
        except Exception as e:
            raise SqlApiUnknownError("error in column list step") from e

    def execute_step(self, step, steps_data):
        if type(step) == GetPredictorColumns:
            predictor_name = step.predictor.parts[-1]
            dn = self.datahub.get(self.mindsdb_database_name)
            columns = dn.get_table_columns(predictor_name)
            columns = [
                (column_name, column_name) for column_name in columns
            ]
            data = {
                'values': [],
                'columns': {
                    (self.mindsdb_database_name, predictor_name, predictor_name): columns
                },
                'tables': [(self.mindsdb_database_name, predictor_name, predictor_name)]
            }
        elif type(step) == GetTableColumns:
            table = step.table
            dn = self.datahub.get(step.namespace)
            ds_query = Select(from_table=Identifier(table), targets=[Star()])

            data, columns_info = dn.query(ds_query)

            table_alias = (self.database, table, table)

            data = {
                'values': [],
                'columns': {
                    table_alias: columns_info
                },
                'tables': [table_alias]
            }
        elif type(step) == FetchDataframeStep:
            data = self._fetch_dataframe_step(step)
        elif type(step) == UnionStep:
            raise ErNotSupportedYet('Union step is not implemented')
            # TODO add union support
            # left_data = steps_data[step.left.step_num]
            # right_data = steps_data[step.right.step_num]
            # data = left_data + right_data
        elif type(step) == MapReduceStep:
            try:
                if step.reduce != 'union':
                    raise ErLogicError(f'Unknown MapReduceStep type: {step.reduce}')

                step_data = steps_data[step.values.step_num]
                vars = []
                step_data_values = step_data['values']
                for row in step_data_values:
                    var_group = {}
                    vars.append(var_group)
                    for row_data in row.values():
                        for name, value in row_data.items():
                            if name[0] != '__mindsdb_row_id':
                                var_group[name[1] or name[0]] = value

                data = {
                    'values': [],
                    'columns': {},
                    'tables': []
                }
                substep = step.step
                if type(substep) == FetchDataframeStep:
                    query = substep.query
                    for var_group in vars:
                        markQueryVar(query.where)
                        for name, value in var_group.items():
                            replaceQueryVar(query.where, value, name)
                        sub_data = self._fetch_dataframe_step(substep)
                        if len(data['columns']) == 0:
                            data['columns'] = sub_data['columns']
                        if len(data['tables']) == 0:
                            data['tables'] = sub_data['tables']
                        data['values'].extend(sub_data['values'])
                        unmarkQueryVar(query.where)
                elif type(substep) == MultipleSteps:
                    data = self._multiple_steps_reduce(substep, vars)
                else:
                    raise ErLogicError(f'Unknown step type: {step.step}')
            except Exception as e:
                raise SqlApiUnknownError(f'error in map reduce step: {e}') from e
        elif type(step) == MultipleSteps:
            if step.reduce != 'union':
                raise ErNotSupportedYet(f"Only MultipleSteps with type = 'union' is supported. Got '{step.type}'")
            data = None
            for substep in step.steps:
                subdata = self.execute_step(substep, steps_data)
                if data is None:
                    data = subdata
                else:
                    data['values'].extend(subdata['values'])
        elif type(step) == ApplyPredictorRowStep:
            try:
                predictor = '.'.join(step.predictor.parts)
                dn = self.datahub.get(self.mindsdb_database_name)
                where_data = step.row_dict

                data = dn.query(
                    table=predictor,
                    columns=None,
                    where_data=where_data,
                    integration_name=self.session.integration,
                    integration_type=self.session.integration_type
                )

                data = [{(key, key): value for key, value in row.items()} for row in data]

                table_name = get_preditor_alias(step, self.database)
                values = [{table_name: x} for x in data]
                columns = {table_name: []}
                if len(data) > 0:
                    row = data[0]
                    columns[table_name] = list(row.keys())
                # TODO else

                data = {
                    'values': values,
                    'columns': columns,
                    'tables': [table_name]
                }
            except Exception as e:
                if isinstance(e, SqlApiException):
                    raise e
                else:
                    raise SqlApiUnknownError(f'error in apply predictor row step: {e}') from e
        elif type(step) in (ApplyPredictorStep, ApplyTimeseriesPredictorStep):
            try:
                dn = self.datahub.get(self.mindsdb_database_name)
                predictor = '.'.join(step.predictor.parts)
                where_data = []
                for row in steps_data[step.dataframe.step_num]['values']:
                    new_row = {}
                    for table_name in row:
                        keys_intersection = set(new_row) & set(row[table_name])
                        if len(keys_intersection) > 0:
                            raise ErLogicError(
                                f'The predictor got two identical keys from different datasources: {keys_intersection}'
                            )
                        new_row.update(row[table_name])
                    where_data.append(new_row)

                where_data = [{key[1]: value for key, value in row.items()} for row in where_data]

                is_timeseries = self.planner.predictor_metadata[predictor]['timeseries']
                _mdb_forecast_offset = None
                if is_timeseries:
                    if '> LATEST' in self.query_str:
                        # stream mode -- if > LATEST, forecast starts on inferred next timestamp
                        _mdb_forecast_offset = 1
                    elif '= LATEST' in self.query_str:
                        # override: when = LATEST, forecast starts on last provided timestamp instead of inferred next time
                        _mdb_forecast_offset = 0
                    else:
                        # normal mode -- emit a forecast ($HORIZON data points on each) for each provided timestamp
                        _mdb_forecast_offset = None
                    for row in where_data:
                        if '__mdb_forecast_offset' not in row:
                            row['__mdb_forecast_offset'] = _mdb_forecast_offset

                for row in where_data:
                    for key in row:
                        if isinstance(row[key], datetime.date):
                            row[key] = str(row[key])

                table_name = get_preditor_alias(step, self.database)
                columns = {table_name: []}
                if len(where_data) == 0:
                    # no data, don't run predictor
                    cols = dn.get_table_columns(predictor) + ['__mindsdb_row_id']
                    columns[table_name] = [(c, c) for c in cols]
                    values = []
                else:
                    data = dn.query(
                        table=predictor,
                        columns=None,
                        where_data=where_data,
                        integration_name=self.session.integration,
                        integration_type=self.session.integration_type
                    )

                    data = [{(key, key): value for key, value in row.items()} for row in data]

                    values = [{table_name: x} for x in data]

                    if len(data) > 0:
                        row = data[0]
                        columns[table_name] = list(row.keys())
                    # TODO else

                data = {
                    'values': values,
                    'columns': columns,
                    'tables': [table_name],
                    'types': {table_name: self.model_types}
                }
            except Exception as e:
                raise SqlApiUnknownError(f'error in apply predictor step: {e}') from e
        elif type(step) == JoinStep:
            try:
                left_data = steps_data[step.left.step_num]
                right_data = steps_data[step.right.step_num]

                # FIXME https://github.com/mindsdb/mindsdb_sql/issues/136
                # is_timeseries = False
                # if True in [type(step) == ApplyTimeseriesPredictorStep for step in plan.steps]:
                #     right_data = steps_data[step.left.step_num]
                #     left_data = steps_data[step.right.step_num]
                #     is_timeseries = True

                if step.query.condition is not None:
                    raise ErNotSupportedYet('At this moment supported only JOIN without condition')
                if step.query.join_type.upper() not in ('LEFT JOIN', 'JOIN'):
                    raise ErNotSupportedYet('At this moment supported only JOIN and LEFT JOIN')
                if (
                        len(left_data['tables']) != 1 or len(right_data['tables']) != 1
                        or left_data['tables'][0] == right_data['tables'][0]
                ):
                    raise ErNotSupportedYet('At this moment supported only JOIN of two different tables')

                data = {
                    'values': [],
                    'columns': {},
                    'tables': list(set(left_data['tables'] + right_data['tables'])),
                    'types': {}
                }

                for data_part in [left_data, right_data]:
                    for table_name in data_part['columns']:
                        if table_name not in data['columns']:
                            data['columns'][table_name] = data_part['columns'][table_name]
                            # keep types
                            data['types'][table_name] = data_part.get('types', {}).get(table_name, {}).copy()
                        else:
                            data['columns'][table_name].extend(data_part['columns'][table_name])
                            # keep types
                            data['types'][table_name].update(data_part.get('types', {}).get(table_name, {}))
                for table_name in data['columns']:
                    data['columns'][table_name] = list(set(data['columns'][table_name]))

                left_key = left_data['tables'][0]
                right_key = right_data['tables'][0]

                left_columns_map = OrderedDict()
                left_columns_map_reverse = OrderedDict()
                for i, column_name in enumerate(left_data['columns'][left_key]):
                    left_columns_map[f'a{i}'] = column_name
                    left_columns_map_reverse[column_name] = f'a{i}'

                right_columns_map = {}
                right_columns_map_reverse = {}
                for i, column_name in enumerate(right_data['columns'][right_key]):
                    right_columns_map[f'b{i}'] = column_name
                    right_columns_map_reverse[column_name] = f'b{i}'

                left_df_data = []
                for row in left_data['values']:
                    row = row[left_key]
                    left_df_data.append({left_columns_map_reverse[key]: value for key, value in row.items()})

                right_df_data = []
                for row in right_data['values']:
                    row = row[right_key]
                    right_df_data.append({right_columns_map_reverse[key]: value for key, value in row.items()})

                df_a = pd.DataFrame(left_df_data, columns=left_columns_map.keys())
                df_b = pd.DataFrame(right_df_data, columns=right_columns_map.keys())

                a_name = f'a{round(time.time() * 1000)}'
                b_name = f'b{round(time.time() * 1000)}'
                con = duckdb.connect(database=':memory:')
                con.register(a_name, df_a)
                con.register(b_name, df_b)
                resp_df = con.execute(f"""
                    SELECT * FROM {a_name} as ta full join {b_name} as tb
                    ON ta.{left_columns_map_reverse[('__mindsdb_row_id', '__mindsdb_row_id')]}
                     = tb.{right_columns_map_reverse[('__mindsdb_row_id', '__mindsdb_row_id')]}
                """).fetchdf()
                con.unregister(a_name)
                con.unregister(b_name)
                con.close()

                resp_df = resp_df.replace({np.nan: None})
                resp_dict = resp_df.to_dict(orient='records')

                for row in resp_dict:
                    new_row = {left_key: {}, right_key: {}}
                    for key, value in row.items():
                        if key.startswith('a'):
                            new_row[left_key][left_columns_map[key]] = value
                        else:
                            new_row[right_key][right_columns_map[key]] = value
                    data['values'].append(new_row)

                # remove all records with empty data from predictor from join result
                # otherwise there are emtpy records in the final result:
                # +------------+------------+-------+-----------+----------+
                # | time       | time       | state | pnew_case | new_case |
                # +------------+------------+-------+-----------+----------+
                # | 2020-10-21 | 2020-10-24 | CA    | 0.0       | 5945.0   |
                # | 2020-10-22 | 2020-10-23 | CA    | 0.0       | 6141.0   |
                # | 2020-10-23 | 2020-10-22 | CA    | 0.0       | 2940.0   |
                # | 2020-10-24 | 2020-10-21 | CA    | 0.0       | 3707.0   |
                # | NULL       | 2020-10-20 | NULL  | nan       | nan      |
                # | NULL       | 2020-10-19 | NULL  | nan       | nan      |
                # | NULL       | 2020-10-18 | NULL  | nan       | nan      |
                # | NULL       | 2020-10-17 | NULL  | nan       | nan      |
                # | NULL       | 2020-10-16 | NULL  | nan       | nan      |
                # +------------+------------+-------+-----------+----------+
                # 9 rows in set (2.07 sec)

                # if is_timeseries:
                #     data_values = []
                #     for row in data['values']:
                #         for key in row:
                #             if 'mindsdb' in key:
                #                 if not is_empty_prediction_row(row[key]):
                #                     data_values.append(row)
                #                     break
                #     data['values'] = data_values
            except Exception as e:
                raise SqlApiUnknownError(f'error in join step: {e}') from e

        elif type(step) == FilterStep:
            step_data = steps_data[step.dataframe.step_num]

            # dicts to look up column and table
            column_idx = {}
            tables_idx = {}
            col_table_idx = {}

            # prepare columns for dataframe. column name contains table name
            cols = set()
            for table, col_list in step_data['columns'].items():
                _, t_name, t_alias = table

                tables_idx[t_name] = t_name
                tables_idx[t_alias] = t_name
                for column in col_list:
                    # table_column
                    c_name, c_alias = column

                    col_name = f'{t_name}^{c_name}'
                    cols.add(col_name)

                    col_table_idx[col_name] = (table, column)
                    column_idx[c_name] = t_name

            # prepare dict for dataframe
            result = []
            for row in step_data['values']:
                data_row = {}
                for table, col_list in step_data['columns'].items():
                    for col in col_list:
                        col_name = f'{table[1]}^{col[0]}'
                        data_row[col_name] = row[table][col]
                result.append(data_row)

            df = pd.DataFrame(result, columns=list(cols))

            # analyze condition and change name of columns
            def check_fields(node, is_table=None, **kwargs):
                if is_table:
                    raise ErNotSupportedYet('Subqueries is not supported in WHERE')
                if isinstance(node, Identifier):
                    # only column name
                    col_name = node.parts[-1]

                    if len(node.parts) == 1:
                        if col_name not in column_idx:
                            raise ErKeyColumnDoesNotExist(f'Table not found for column: {col_name}')
                        table_name = column_idx[col_name]
                    else:
                        table_name = node.parts[-2]

                        # maybe it is alias
                        table_name = tables_idx[table_name]

                    new_name = f'{table_name}^{col_name}'
                    return Identifier(parts=[new_name])

            where_query = step.query
            query_traversal(where_query, check_fields)

            query = Select(targets=[Star()], from_table=Identifier('df'), where=where_query)

            res = query_df(df, query)

            resp_dict = res.to_dict(orient='records')

            # convert result to dict-dict structure
            values = []
            for row in resp_dict:
                value = {}
                for key, v in row.items():
                    # find original table and col
                    table, column = col_table_idx[key]
                    if table not in value:
                        value[table] = {}
                    value[table][column] = v

                values.append(value)

            data = {
                'tables': step_data['tables'],
                'columns': step_data['columns'],
                'values': values,
                'types': step_data.get('types', {})
            }

        elif type(step) == LimitOffsetStep:
            try:
                step_data = steps_data[step.dataframe.step_num]
                data = {
                    'values': step_data['values'].copy(),
                    'columns': step_data['columns'].copy(),
                    'tables': step_data['tables'].copy(),
                    'types': step_data.get('types', {}).copy(),
                }
                if isinstance(step.offset, Constant) and isinstance(step.offset.value, int):
                    data['values'] = data['values'][step.offset.value:]
                if isinstance(step.limit, Constant) and isinstance(step.limit.value, int):
                    data['values'] = data['values'][:step.limit.value]
            except Exception as e:
                raise SqlApiUnknownError(f'error in limit offset step: {e}') from e
        elif type(step) == ProjectStep:
            try:
                step_data = steps_data[step.dataframe.step_num]

                columns = defaultdict(list)
                for column_identifier in step.columns:

                    if type(column_identifier) == Star:
                        for table_name, table_columns_list in step_data['columns'].items():
                            for column in table_columns_list:

                                columns[table_name].append(column)

                    elif type(column_identifier) == Identifier:
                        appropriate_table = None
                        columns_to_copy = None

                        column_name_parts = column_identifier.parts
                        column_alias = column_identifier.parts[-1] if column_identifier.alias is None else '.'.join(
                            column_identifier.alias.parts)
                        if len(column_name_parts) > 2:
                            raise ErSqlWrongArguments(
                                f'Column name must contain no more than 2 parts. Got name: {column_identifier}')
                        elif len(column_name_parts) == 1:
                            column_name = column_name_parts[0]

                            for table_name, table_columns in step_data['columns'].items():
                                table_col_idx = {}
                                for x in table_columns:
                                    name = x[1] or x[0]
                                    table_col_idx[name] = x

                                column_exists = get_column_in_case(list(table_col_idx.keys()), column_name)
                                if column_exists:
                                    if appropriate_table is not None and not step.ignore_doubles:
                                        raise ErLogicError(
                                            f'Found multiple appropriate tables for column {column_name}')
                                    else:
                                        appropriate_table = table_name
                                        new_col = (column_name, column_alias)
                                        cur_col = table_col_idx[column_exists]

                                        columns[appropriate_table].append(new_col)
                                        if cur_col != new_col:
                                            columns_to_copy = cur_col, new_col
                                        break

                            if appropriate_table is None:
                                raise SqlApiException(f'Can not find appropriate table for column {column_name}')

                        elif len(column_name_parts) == 2:
                            table_name_or_alias = column_name_parts[0]
                            column_name = column_name_parts[1]

                            for table_name, table_columns in step_data['columns'].items():
                                checking_table_name_or_alias = table_name[2] or table_name[1]
                                if table_name_or_alias.lower() == checking_table_name_or_alias.lower():
                                    # support select table.*
                                    if isinstance(column_name, Star):
                                        # add all by table
                                        appropriate_table = table_name
                                        for column in step_data['columns'][appropriate_table]:
                                            columns[appropriate_table].append(column)
                                        break

                                    table_col_idx = {}
                                    for x in table_columns:
                                        name = x[1] or x[0]
                                        table_col_idx[name] = x

                                    column_exists = get_column_in_case(list(table_col_idx.keys()), column_name)
                                    if column_exists:
                                        appropriate_table = table_name
                                        new_col = (column_name, column_alias)
                                        cur_col = table_col_idx[column_exists]

                                        columns[appropriate_table].append(new_col)
                                        if cur_col != new_col:
                                            columns_to_copy = cur_col, new_col

                                        break
                                    else:
                                        raise ErLogicError(f'Can not find column "{column_name}" in table "{table_name}"')
                            if appropriate_table is None:
                                raise ErLogicError(f'Can not find appropriate table for column {column_name}')
                        else:
                            raise ErSqlWrongArguments('Undefined column name')

                        if columns_to_copy is not None:
                            col_from, col_to = columns_to_copy
                            for row in step_data['values']:
                                row[appropriate_table][col_to] = row[appropriate_table][col_from]
                            # TODO copy types?

                    else:
                        raise ErKeyColumnDoesNotExist(f'Unknown column type: {column_identifier}')

                data = {
                    'tables': step_data['tables'],
                    'columns': columns,
                    'values': step_data['values'],  # TODO keep values only for columns
                    'types': step_data.get('types', {})
                }
            except Exception as e:
                if isinstance(e, SqlApiException):
                    raise e
                raise SqlApiUnknownError(f'error on project step: {e} ') from e
        elif type(step) == GroupByStep:
            step_data = steps_data[step.dataframe.step_num]

            result = []
            cols = set()
            for _, col_list in step_data['columns'].items():
                for col in col_list:
                    cols.add(col[0])

            for row in step_data['values']:
                data_row = {}
                for table, col_list in step_data['columns'].items():
                    for col in col_list:
                        data_row[col[0]] = row[table][col]
                result.append(data_row)

            df = pd.DataFrame(result, columns=list(cols))

            query = Select(targets=step.targets, from_table='df', group_by=step.columns).to_string()
            res = query_df(df, query)

            resp_dict = res.to_dict(orient='records')

            # stick all columns to first table
            appropriate_table = step_data['tables'][0]

            columns_list = []
            columns = []
            for key, dtyp in res.dtypes.items():
                columns_list.append(
                    Column(database=appropriate_table[0],
                           table_name=appropriate_table[1],
                           table_alias=appropriate_table[2],
                           name=key,
                           type=dtyp,
                           alias=key)
                )
                columns.append((key, key))

            values = []
            for row in resp_dict:
                row2 = {
                    (key, key): value
                    for key, value in row.items()
                }

                values.append(
                    {
                        appropriate_table: row2
                    }
                )

            # columns are changed
            self.columns_list = columns_list
            types = step_data.get('types', {})
            data = {
                'tables': [appropriate_table],
                'columns': {appropriate_table: columns},
                'values': values,
                'types': types  # copy
            }

        elif type(step) == SaveToTable or type(step) == InsertToTable:
            is_replace = False
            is_create = False

            if type(step) == SaveToTable:
                is_create = True

                if step.is_replace:
                    is_replace = True

            step_data = step.dataframe.result_data
            integration_name = step.table.parts[0]
            table_name_parts = step.table.parts[1:]

            dn = self.datahub.get(integration_name)

            if hasattr(dn, 'create_table') is False:
                raise ErNotSupportedYet(f"Creating table in '{integration_name}' is not supporting")

            # region del 'service' columns
            for table in step_data['columns']:
                new_table_columns = []
                for column in step_data['columns'][table]:
                    if column[-1] not in ('__mindsdb_row_id', '__mdb_forecast_offset'):
                        new_table_columns.append(column)
                step_data['columns'][table] = new_table_columns
            # endregion

            # region del columns filtered at projection step
            if self.columns_list is not None:
                filtered_column_names = [x.name for x in self.columns_list]
                for table in step_data['columns']:
                    new_table_columns = []
                    for column in step_data['columns'][table]:
                        if column[0].startswith('predictor.'):
                            new_table_columns.append(column)
                        elif column[0] in filtered_column_names:
                            new_table_columns.append(column)
                    step_data['columns'][table] = new_table_columns
            # endregion

            # drop double names
            if len(step_data['tables']) > 1:
                # set prefixes for all tables except first one
                for table in step_data['tables'][1:]:
                    table_name = table[1]
                    col_map = []
                    col_list = []
                    for column in step_data['columns'][table]:
                        column_new = (f'{table_name}.{column[0]}', f'{table_name}.{column[1]}')
                        col_list.append(column_new)
                        col_map.append([column, column_new])

                    # replace columns
                    step_data['columns'][table] = col_list

                    # replace in values
                    for row in step_data['values']:
                        table_row = row[table]

                        for column, column_new in col_map:
                            table_row[column_new] = table_row.pop(column)

            dn.create_table(
                table_name_parts=table_name_parts,
                columns=step_data['columns'],
                data=step_data['values'],
                is_replace=is_replace,
                is_create=is_create
            )
            data = None
        else:
            raise ErLogicError(F'Unknown planner step: {step}')
        return data

    # Is not used
    # def _apply_where_filter(self, row, where):
    #     if isinstance(where, Identifier):
    #         return row[where.value]
    #     elif isinstance(where, Constant):
    #         return where.value
    #     elif not isinstance(where, (UnaryOperation, BinaryOperation)):
    #         raise SqlApiException(f'Unknown operation type: {where}')
    #
    #     op_fn = operator_map.get(where.op)
    #     if op_fn is None:
    #         raise SqlApiException(f'unknown operator {where.op}')
    #
    #     args = [self._apply_where_filter(row, arg) for arg in where.args]
    #     result = op_fn(*args)
    #     return result

    def _make_list_result_view(self, data):
        if self.outer_query is not None:
            return data['values']
        result = []
        for row in data['values']:
            data_row = []
            for column_record in self.columns_list:
                table_name = (column_record.database, column_record.table_name, column_record.table_alias)
                column_name = (column_record.name, column_record.alias)
                if table_name in row is False:
                    # try without alias
                    table_name = (table_name[0], table_name[1], None)

                data_row.append(row[table_name][column_name])
            result.append(data_row)
        return result

    def _make_dict_result_view(self, data):
        result = []
        for row in data['values']:
            data_row = {}
            for table_name in row:
                data_row.update(row[table_name])
            result.append(data_row)
        return result



import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes import Responder
from mindsdb.interfaces.storage.db import session
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb.api.mongo.utilities import log
from mindsdb.integrations.libs.response import HandlerStatusResponse

class Responce(Responder):
    when = {'insert': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            res = self._result(query, request_env, mindsdb_env)
        except Exception as e:
            log.error(e)
            res = {
                'n': 0,
                'writeErrors': [{
                    'index': 0,
                    'code': 0,
                    'errmsg': str(e)
                }],
                'ok': 1
            }
        return res

    def _result(self, query, request_env, mindsdb_env):
        table = query['insert']

        if table == 'databases':
            for doc in query['documents']:
                if '_id' in doc:
                    del doc['_id']
                for field in ('name', 'engine', 'connection_args'):
                    if field not in doc:
                        raise Exception(f"'{field}' must be specified")

                status = HandlerStatusResponse(success=False)
                try:
                    handler = mindsdb_env['integration_controller'].create_handler(
                        handler_type=doc['engine'],
                        connection_data=doc['connection_args']
                    )
                    status = handler.check_connection()
                except Exception as e:
                    status.error_message = str(e)

                if status.success is False:
                    raise Exception(f"Can't connect to db: {status.error_message}")

                integration = mindsdb_env['integration_controller'].get(doc['name'])
                if integration is not None:
                    raise Exception(f"Database '{doc['name']}' already exists.")

            for doc in query['documents']:
                mindsdb_env['integration_controller'].add(doc['name'], doc['engine'], doc['connection_args'])

            result = {
                "n": len(query['documents']),
                "ok": 1
            }
        elif table == 'predictors':
            predictors_columns = [
                'name',
                'status',
                'accuracy',
                'predict',
                'select_data_query',
                'training_options',
                'connection'
            ]

            models = mindsdb_env['model_interface'].get_models()

            if len(query['documents']) != 1:
                raise Exception("Must be inserted just one predictor at time")

            for doc in query['documents']:
                if '_id' in doc:
                    del doc['_id']

                bad_columns = [x for x in doc if x not in predictors_columns]
                if len(bad_columns) > 0:
                    raise Exception(f"Is no possible insert this columns to 'predictors' collection: {', '.join(bad_columns)}")

                if 'name' not in doc:
                    raise Exception("Please, specify 'name' field")

                if 'predict' not in doc:
                    raise Exception("Please, specify 'predict' field")

                if doc['name'] in [x['name'] for x in models]:
                    raise Exception(f"Predictor with name '{doc['name']}' already exists")

                select_data_query = doc.get('select_data_query')
                if select_data_query is None:
                    raise Exception("'select_data_query' must be in query")

                kwargs = doc.get('training_options', {})
                if 'timeseries_settings' in kwargs:
                    # mongo shell client sends int as float. need to convert it to int
                    for key in ('window', 'horizon'):
                        val = kwargs['timeseries_settings'].get(key)
                        if val is not None:
                            kwargs['timeseries_settings'][key] = int(val)

                integrations = mindsdb_env['integration_controller'].get_all().keys()
                connection = doc.get('connection')
                if connection is None:
                    if 'default_mongodb' in integrations:
                        connection = 'default_mongodb'
                    else:
                        for integration in integrations:
                            if integration.startswith('mongodb_'):
                                connection = integration
                                break

                if connection is None:
                    raise Exception("Can't find connection for data source")

                handler = mindsdb_env['integration_controller'].get_handler(connection)
                result = handler.native_query(select_data_query)

                if result.type != RESPONSE_TYPE.TABLE:
                    raise Exception(f'Error during query: {result.error_message}')

                ds_data_df = result.data_frame
                ds_column_names = list(ds_data_df.columns)

                predict = doc['predict']
                if not isinstance(predict, list):
                    predict = [x.strip() for x in predict.split(',')]

                for col in predict:
                    if col not in ds_column_names:
                        raise Exception(f"Column '{col}' not exists")

                mindsdb_env['model_interface'].learn(
                    doc['name'], ds_data_df, predict, kwargs=dict(kwargs),
                    user_class=mindsdb_env.get('user_class', 0)
                )

            result = {
                "n": len(query['documents']),
                "ok": 1
            }
        else:
            raise Exception("Only insert to 'predictors' or 'databases' allowed")

        return result


responder = Responce()



import os
import sys
import json

from mindsdb.__about__ import __version__  # noqa: F401
from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.utilities.fs import get_or_create_data_dir, create_dirs_recursive
from mindsdb.utilities.functions import args_parse, is_notebook
from mindsdb.utilities.telemetry import telemetry_file_exists, disable_telemetry

is_ray_worker = False
if sys.argv[0].endswith('ray/workers/default_worker.py'):
    is_ray_worker = True

is_alembic = os.path.basename(sys.argv[0]).split('.')[0] == 'alembic'
is_pytest = os.path.basename(sys.argv[0]).split('.')[0] == 'pytest'

if not is_ray_worker:
    try:
        if not is_notebook() and not is_alembic and not is_pytest:

            args = args_parse()
        else:
            args = None
    except Exception:
        # This fials in some notebooks ... check above for is_notebook is still needed because even if the exception is caught trying to read the arg still leads to failure in other notebooks... notebooks a
        args = None

    # ---- CHECK SYSTEM ----
    if not (sys.version_info[0] >= 3 and sys.version_info[1] >= 6):
        print("""
    MindsDB server requires Python >= 3.7 to run

    Once you have Python 3.7 installed you can tun mindsdb as follows:

    1. create and activate venv:
    python3.7 -m venv venv
    source venv/bin/activate

    2. install MindsDB:
    pip3 install mindsdb

    3. Run MindsDB
    python3.7 -m mindsdb

    More instructions in https://docs.mindsdb.com
        """)
        exit(1)

    # --- VERSION MODE ----
    if args is not None and args.version:
        print(f'MindsDB {mindsdb_version}')
        sys.exit(0)

    # --- MODULE OR LIBRARY IMPORT MODE ----
    if args is not None and args.config is not None:
        config_path = args.config
        with open(config_path, 'r') as fp:
            user_config = json.load(fp)
    else:
        user_config = {}
        config_path = 'absent'
    os.environ['MINDSDB_CONFIG_PATH'] = config_path

    if 'storage_dir' in user_config:
        root_storage_dir = user_config['storage_dir']
        os.environ['MINDSDB_STORAGE_DIR'] = root_storage_dir
    elif os.environ.get('MINDSDB_STORAGE_DIR') is not None:
        root_storage_dir = os.environ['MINDSDB_STORAGE_DIR']
    else:
        root_storage_dir = get_or_create_data_dir()
        os.environ['MINDSDB_STORAGE_DIR'] = root_storage_dir

    if os.path.isdir(root_storage_dir) is False:
        os.makedirs(root_storage_dir)

    if 'storage_db' in user_config:
        os.environ['MINDSDB_DB_CON'] = user_config['storage_db']
    elif os.environ.get('MINDSDB_DB_CON', '') == '':
        os.environ['MINDSDB_DB_CON'] = 'sqlite:///' + os.path.join(os.environ['MINDSDB_STORAGE_DIR'], 'mindsdb.sqlite3.db') + '?check_same_thread=False&timeout=30'

    from mindsdb.utilities.config import Config
    mindsdb_config = Config()
    create_dirs_recursive(mindsdb_config['paths'])

    os.environ['DEFAULT_LOG_LEVEL'] = os.environ.get('DEFAULT_LOG_LEVEL', 'ERROR')
    os.environ['LIGHTWOOD_LOG_LEVEL'] = os.environ.get('LIGHTWOOD_LOG_LEVEL', 'ERROR')
    os.environ['MINDSDB_STORAGE_PATH'] = mindsdb_config['paths']['predictors']

    if telemetry_file_exists(mindsdb_config['storage_dir']):
        os.environ['CHECK_FOR_UPDATES'] = '0'
        print('\n x telemetry disabled! \n')
    elif os.getenv('CHECK_FOR_UPDATES', '1').lower() in ['0', 'false', 'False'] or mindsdb_config.get('cloud', False):
        disable_telemetry(mindsdb_config['storage_dir'])
        print('\n x telemetry disabled \n')
    else:
        print('\n ✓ telemetry enabled \n')



from io import BytesIO, StringIO
import os
import csv
import json
import codecs
import traceback
import tempfile
from urllib.parse import urlparse

import requests
import pandas as pd

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import DropTables, Select

from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.integrations.libs.base_handler import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


def clean_row(row):
    n_row = []
    for cell in row:
        if str(cell) in ['', ' ', '  ', 'NaN', 'nan', 'NA']:
            n_row.append(None)
        else:
            n_row.append(cell)

    return n_row


class FileHandler(DatabaseHandler):
    """
    Handler for files
    """
    name = 'files'

    def __init__(self, name=None, db_store=None, fs_store=None, connection_data=None, file_controller=None):
        super().__init__(name)
        self.parser = parse_sql
        self.fs_store = fs_store
        self.custom_parser = connection_data.get('custom_parser')
        self.clean_rows = connection_data.get('clean_rows', True)
        self.file_controller = file_controller

    def connect(self, **kwargs):
        return

    def disconnect(self, **kwargs):
        return

    def check_connection(self) -> StatusResponse:
        return StatusResponse(True)

    def query(self, query: ASTNode) -> Response:
        if type(query) == DropTables:
            for table_identifier in query.tables:
                if len(table_identifier.parts) == 2 and table_identifier.parts[0] != self.name:
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table from database '{table_identifier.parts[0]}'"
                    )
                table_name = table_identifier.parts[-1]
                try:
                    self.file_controller.delete_file(table_name)
                except Exception as e:
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table '{table_name}': {e}"
                    )
            return Response(RESPONSE_TYPE.OK)
        elif type(query) == Select:
            table_name = query.from_table.parts[-1]
            file_path = self.file_controller.get_file_path(table_name)
            df, _columns = self._handle_source(file_path, self.clean_rows, self.custom_parser)
            result_df = query_df(df, query)
            return Response(
                RESPONSE_TYPE.TABLE,
                data_frame=result_df
            )
        else:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message="Only 'select' and 'drop' queries allowed for files"
            )

    def native_query(self, query: str) -> Response:
        ast = self.parser(query, dialect='mindsdb')
        return self.query(ast)

    @staticmethod
    def _handle_source(file_path, clean_rows=True, custom_parser=None):
        # get file data io, format and dialect
        data, fmt, dialect = FileHandler._get_data_io(file_path)
        data.seek(0)  # make sure we are at 0 in file pointer

        if custom_parser:
            header, file_data = custom_parser(data, fmt)

        elif fmt == 'csv':
            csv_reader = list(csv.reader(data, dialect))
            header = csv_reader[0]
            file_data = csv_reader[1:]

        elif fmt in ['xlsx', 'xls']:
            data.seek(0)
            df = pd.read_excel(data)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        elif fmt == 'json':
            data.seek(0)
            json_doc = json.loads(data.read())
            df = pd.json_normalize(json_doc, max_level=0)
            header = df.columns.values.tolist()
            file_data = df.values.tolist()

        else:
            raise ValueError('Could not load file into any format, supported formats are csv, json, xls, xlsx')

        if clean_rows:
            file_list_data = [clean_row(row) for row in file_data]
        else:
            file_list_data = file_data

        header = [x.strip() for x in header]
        col_map = dict((col, col) for col in header)
        return pd.DataFrame(file_list_data, columns=header), col_map

    @staticmethod
    def _get_data_io(file_path):
        """
        This gets a file either url or local file and defiens what the format is as well as dialect
        :param file: file path or url
        :return: data_io, format, dialect
        """

        ############
        # get file as io object
        ############

        # file_path = self._get_file_path()

        data = BytesIO()

        try:
            with open(file_path, 'rb') as fp:
                data = BytesIO(fp.read())
        except Exception as e:
            error = 'Could not load file, possible exception : {exception}'.format(exception=e)
            print(error)
            raise ValueError(error)

        dialect = None

        ############
        # check for file type
        ############

        # try to guess if its an excel file
        xlsx_sig = b'\x50\x4B\x05\06'
        # xlsx_sig2 = b'\x50\x4B\x03\x04'
        xls_sig = b'\x09\x08\x10\x00\x00\x06\x05\x00'

        # different whence, offset, size for different types
        excel_meta = [('xls', 0, 512, 8), ('xlsx', 2, -22, 4)]

        for filename, whence, offset, size in excel_meta:

            try:
                data.seek(offset, whence)  # Seek to the offset.
                bytes = data.read(size)  # Capture the specified number of bytes.
                data.seek(0)
                codecs.getencoder('hex')(bytes)

                if bytes == xls_sig:
                    return data, 'xls', dialect
                elif bytes == xlsx_sig:
                    return data, 'xlsx', dialect

            except Exception:
                data.seek(0)

        # if not excel it can be a json file or a CSV, convert from binary to stringio

        byte_str = data.read()
        # Move it to StringIO
        try:
            # Handle Microsoft's BOM "special" UTF-8 encoding
            if byte_str.startswith(codecs.BOM_UTF8):
                data = StringIO(byte_str.decode('utf-8-sig'))
            else:
                data = StringIO(byte_str.decode('utf-8'))

        except Exception:
            print(traceback.format_exc())
            print('Could not load into string')

        # see if its JSON
        buffer = data.read(100)
        data.seek(0)
        text = buffer.strip()
        # analyze first n characters
        if len(text) > 0:
            text = text.strip()
            # it it looks like a json, then try to parse it
            if text.startswith('{') or text.startswith('['):
                try:
                    json.loads(data.read())
                    data.seek(0)
                    return data, 'json', dialect
                except Exception:
                    data.seek(0)
                    return data, None, dialect

        # lets try to figure out if its a csv
        try:
            dialect = FileHandler._get_csv_dialect(file_path)
            if dialect:
                return data, 'csv', dialect
            return data, None, dialect
        except Exception:
            data.seek(0)
            print('Could not detect format for this file')
            print(traceback.format_exc())
            # No file type identified
            return data, None, dialect

    @staticmethod
    def _get_file_path(path) -> str:
        try:
            is_url = urlparse(path).scheme in ('http', 'https')
        except Exception:
            is_url = False
        if is_url:
            path = FileHandler._fetch_url(path)
        return path

    @staticmethod
    def _get_csv_dialect(file_path) -> csv.Dialect:
        with open(file_path, 'rt') as f:
            try:
                accepted_csv_delimiters = [',', '\t', ';']
                dialect = csv.Sniffer().sniff(f.read(128 * 1024), delimiters=accepted_csv_delimiters)
            except csv.Error:
                dialect = None
        return dialect

    @staticmethod
    def _fetch_url(url: str) -> str:
        temp_dir = tempfile.mkdtemp(prefix='mindsdb_file_url_')
        try:
            r = requests.get(url, stream=True)
            if r.status_code == 200:
                with open(os.path.join(temp_dir, 'file'), 'wb') as f:
                    for chunk in r:
                        f.write(chunk)
            else:
                raise Exception(f'Responce status code is {r.status_code}')
        except Exception as e:
            print(f'Error during getting {url}')
            print(e)
            raise
        return os.path.join(temp_dir, 'file')

    def get_tables(self) -> Response:
        """
        List all files
        """
        files_meta = self.file_controller.get_files()
        data = [{
            'TABLE_NAME': x['name'],
            'TABLE_ROWS': x['row_count']
        } for x in files_meta]
        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(data)
        )

    def get_columns(self, table_name) -> Response:
        file_meta = self.file_controller.get_file_meta(table_name)
        result = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame([
                {
                    'Field': x.strip(),
                    'Type': 'str'
                } for x in file_meta['columns']
            ])
        )
        return result



from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.mongo.classes import Responder


class Responce(Responder):
    def when(self, query):
        return (
            'buildinfo' in query or 'buildInfo' in query
        )

    result = {
        # 'version': f'MindsDB {mindsdb_version}',
        'version': '3.6',
        'ok': 1
    }


responder = Responce()



import unittest
from mindsdb.integrations.handlers.mssql_handler.mssql_handler import SqlServerHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class SqlServerHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "localhost",
            "port": "1433",
            "user": "sa",
            "password": "",
            "database": "master"
        }
        cls.handler = SqlServerHandler('test_sqlserver_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_describe_table(self):
        described = self.handler.describe_table("home_rentals")
        assert described['type'] is not RESPONSE_TYPE.ERROR

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables['type'] is not RESPONSE_TYPE.ERROR

    def test_3_get_views(self):
        views = self.handler.get_views()
        assert views['type'] is not RESPONSE_TYPE.ERROR

    def test_4_select_query(self):
        query = "SELECT * FROM test_data.home_rentals"
        result = self.handler.native_query(query)
        assert result['type'] is RESPONSE_TYPE.TABLE


from typing import Optional, Any

import pandas as pd
from mindsdb_sql.parser.ast import Join
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.response import HandlerResponse, HandlerStatusResponse


class BaseHandler:
    """ Base class for handlers

    Base class for handlers that associate a source of information with the
    broader MindsDB ecosystem via SQL commands.
    """

    def __init__(self, name: str):
        """ constructor
        Args:
            name (str): the handler name
        """
        self.is_connected: bool = False
        self.name = name

    def connect(self) -> HandlerStatusResponse:
        """ Set up any connections required by the handler

        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.

        Returns:
            HandlerStatusResponse
        """
        raise NotImplementedError()

    def disconnect(self):
        """ Close any existing connections

        Should switch self.is_connected.
        """
        self.is_connected = False
        return

    def check_connection(self) -> HandlerStatusResponse:
        """ Cehck connection to the handler

        Returns:
            HandlerStatusResponse
        """
        raise NotImplementedError()

    def native_query(self, query: Any) -> HandlerResponse:
        """Receive raw query and act upon it somehow.

        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def query(self, query: ASTNode) -> HandlerResponse:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def get_tables(self) -> HandlerResponse:
        """ Return list of entities

        Return list of entities that will be accesible as tables.

        Returns:
            HandlerResponse: shoud have same columns as information_schema.tables
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-tables-table.html)
                Column 'TABLE_NAME' is mandatory, other is optional.
        """
        raise NotImplementedError()

    def get_columns(self, table_name: str) -> HandlerResponse:
        """ Returns a list of entity columns

        Args:
            table_name (str): name of one of tables returned by self.get_tables()

        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """
        raise NotImplementedError()


class DatabaseHandler(BaseHandler):
    """
    Base class for handlers associated to data storage systems (e.g. databases, data warehouses, streaming services, etc.)
    """
    def __init__(self, name: str):
        super().__init__(name)


class PredictiveHandler(BaseHandler):
    """
    Base class for handlers associated to predictive systems.
    """
    def __init__(self, name: str):
        super().__init__(name)

    def join(self, stmt, data_handler, into: Optional[str]) -> pd.DataFrame:
        """
        Join the output of some entity in the handler with output from some other handler.

        Data from the external handler should be retrieved via the `select_query` method.

        `into`: if provided, the resulting output will be stored in the specified data handler table via `handler.select_into()`. 
        """
        raise NotImplementedError()

    def _get_model_name(self, stmt):
        """ Discern between joined entities to retrieve model name, alias and the clause side it is on. """
        side = None
        models = self.get_tables().data_frame['model_name'].values
        if type(stmt.from_table) == Join:
            model_name = stmt.from_table.right.parts[-1]
            side = 'right'
            if model_name not in models:
                model_name = stmt.from_table.left.parts[-1]
                side = 'left'
            alias = str(getattr(stmt.from_table, side).alias)
        else:
            model_name = stmt.from_table.parts[-1]
            alias = None  # todo: fix this

        if model_name not in models:
            raise Exception("Error, not found. Please create this predictor first.")

        return model_name, alias, side



from setuptools import setup, find_packages


about = {}
with open("mindsdb/__about__.py") as fp:
    exec(fp.read(), about)


with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()


def install_deps():
    """Reads requirements.txt and preprocess it
    to be feed into setuptools.

    This is the only possible way (we found)
    how requirements.txt can be reused in setup.py
    using dependencies from private github repositories.

    Links must be appendend by `-{StringWithAtLeastOneNumber}`
    or something like that, so e.g. `-9231` works as well as
    `1.1.0`. This is ignored by the setuptools, but has to be there.

    Warnings:
        to make pip respect the links, you have to use
        `--process-dependency-links` switch. So e.g.:
        `pip install --process-dependency-links {git-url}`

    Returns:
         list of packages and dependency links.
    """
    default = open('requirements.txt', 'r').readlines()
    new_pkgs = []
    links = []
    for resource in default:
        if 'git+https' in resource:
            pkg = resource.split('#')[-1]
            links.append(resource.strip() + '-9876543210')
            new_pkgs.append(pkg.replace('egg=', '').rstrip())
        else:
            new_pkgs.append(resource.strip())
    return new_pkgs, links


pkgs, new_links = install_deps()

setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__github__'],
    download_url=about['__pypi__'],
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__email__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=pkgs,
    dependency_links=new_links,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7"
)



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.datahub import init_datahub
from mindsdb.api.mysql.mysql_proxy.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.with_kwargs_wrapper import WithKWArgsWrapper


class SessionController():
    '''
    This class manages the server session
    '''

    def __init__(self, server, company_id: int = None, user_class: int = None) -> object:
        """
        Initialize the session
        :param company_id:
        """

        self.username = None
        self.user_class = user_class
        self.auth = False
        self.company_id = company_id
        self.logging = log

        self.integration = None
        self.integration_type = None
        self.database = None

        self.config = Config()

        self.model_interface = WithKWArgsWrapper(
            server.original_model_interface,
            company_id=company_id
        )

        self.integration_controller = WithKWArgsWrapper(
            server.original_integration_controller,
            company_id=company_id
        )

        self.view_interface = WithKWArgsWrapper(
            server.original_view_controller,
            company_id=company_id
        )

        self.datahub = init_datahub(self)

        self.prepared_stmts = {}
        self.packet_sequence_number = 0

    def inc_packet_sequence_number(self):
        self.packet_sequence_number = (self.packet_sequence_number + 1) % 256

    def register_stmt(self, statement):
        i = 1
        while i in self.prepared_stmts and i < 100:
            i = i + 1
        if i == 100:
            raise Exception('Too many unclosed queries')

        self.prepared_stmts[i] = dict(
            type=None,
            statement=statement,
            fetched=0
        )
        return i

    def unregister_stmt(self, stmt_id):
        del self.prepared_stmts[stmt_id]



from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from flask_restx import fields
from collections import OrderedDict

predictor_query_params = OrderedDict([
    ('name', {
        'description': 'The predictor name',
        'type': 'string',
        'in': 'path',
        'required': True
    }),
    ('when', {
        'description': 'The query conditions: key - column, value - value',
        'type': 'object',
        'in': 'body',
        'required': True,
        'example': "{'number_of_rooms': 2,'number_of_bathrooms':1, 'sqft': 1190}"
    }),
])

upload_predictor_params = OrderedDict([
    ('file', {
        'description': 'file',
        'type': 'file',
        'in': 'FormData',
        'required': True
    })
])

put_predictor_metadata = ns_conf.model('PUTPredictorMetadata', {
    'data_source_name': fields.String(
        required=False,
        description='Datasource name. Outdated, will be removed soon.'
    ),
    'from': fields.Nested(
        ns_conf.model('PUTPredictorMetadata_from', {
            'datasource': fields.String(required=False, description='Name of datasource'),
            'query': fields.String(required=False, description='Query to datasource', )
        }),
        required=False,
        description='Source of data for predictor training'
    ),
    'to_predict': fields.String(
        required=True,
        description='Predicted field name'
    ),
    'kwargs': fields.Raw(default={})
})

put_predictor_params = OrderedDict([
    ('name', {
        'description': 'The predictor name',
        'type': 'string',
        'in': 'path',
        'required': True
    }),
    ('data_source_name', {
        'description': 'The data source name',
        'type': 'string',
        'in': 'body',
        'required': False
    }),
    ('datasource', {
        'description': '',
        'type': 'string',
        'in': 'body',
        'required': False,
        'example': '{"name": "mysql_ds", "query": "select * from db.table"}'
    }),
    ('to_predict', {
        'description': 'list of column names to predict',
        'type': 'array',
        'in': 'body',
        'required': True,
        'example': "['number_of_rooms', 'price']"
    }),
    ('kwargs', {
        'description': 'list of column names to predict',
        'type': 'object',
        'in': 'body',
        'required': False,
        'example': '{"advanced_args": {"use_selfaware_model": false}}'
    })
])



import os
import shelve
import json
from abc import ABC, abstractmethod

import walrus
from mindsdb.utilities.config import Config

CONFIG = Config()


class BaseCache(ABC):
    def __init__(self):
        self.config = Config()

    @abstractmethod
    def delete(self):
        pass

    @abstractmethod
    def __getitem__(self, key):
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        pass



class LocalCache(BaseCache):
    def __init__(self, name, *args, **kwargs):
        super().__init__()
        self.kwargs = kwargs
        self.cache_file = os.path.join(self.config['paths']['cache'], name)
        self.cache = shelve.open(self.cache_file, **kwargs)

    def __getattr__(self, name):
        return getattr(self.cache, name)

    def __getitem__(self, key):
        return self.cache.__getitem__(key)

    def __setitem__(self, key, value):
        return self.cache.__setitem__(key, value)

    def __enter__(self):
        if self.cache is None:
            self.cache = shelve.open(self.cache_file, **self.kwargs)
        return self.cache.__enter__()

    def __exit__(self, _type, value, traceback):
        if self.cache is None:
            return None
        res = self.cache.__exit__(_type, value, traceback)
        self.cache = None
        return res

    def __contains__(self, key):
        return key in self.cache

    def delete(self):
        try:
            self.cache.close()
        except Exception:
            pass
        os.remove(self.cache_file)


class RedisCache(BaseCache):
    def __init__(self, prefix, *args, **kwargs):
        super().__init__()
        self.prefix = prefix
        if self.config["cache"]["type"] != "redis":
            raise Exception(f"wrong cache type in config. expected 'redis', but got {self.config['cache']['type']}.")
        connection_info = self.config["cache"]["params"]
        self.client = walrus.Database(**connection_info)

    def __decode(self, data):

        if isinstance(data, dict):
            return dict((self.__decode(x), self.__decode(data[x])) for x in data)
        if isinstance(data, list):
            return list(self.__decode(x) for x in data)
        # assume it is string
        return data.decode("utf8")

    def __contains__(self, key):
        key = f"{self.prefix}_{key}"
        return key in self.__decode(self.client.keys())

    def __getitem__(self, key):
        key = f"{self.prefix}_{key}"
        raw = self.client.get(key)
        if raw is None:
            raise KeyError(key)
        try:
            res = json.loads(raw)
        except json.JSONDecodeError:
            res = raw.decode('utf8')
        return res

    def __setitem__(self, key, value):
        key = f"{self.prefix}_{key}"
        self.client.set(key, json.dumps(value))

    def __iter__(self):
        return iter(self.__decode(self.client.keys()))

    def __next__(self):
        for i in self.__decode(self.client.keys()):
            yield i

    def __delitem__(self, key):
        key = f"{self.prefix}_key"
        self.client.delete(key)

    def delete(self):
        pass


Cache = RedisCache if CONFIG['cache']['type'] == 'redis' else LocalCache
